import base64
import binascii
import inspect
import io
import itertools
import logging
from pathlib import Path
import subprocess
import sys
import struct
import webbrowser
from configparser import ConfigParser
from importlib.machinery import SourceFileLoader
from pprint import pformat
from typing import Any, Dict, List, Mapping, Tuple, Union, Callable
from urllib.parse import urljoin

import lazy_import
import pyperclip
import json
import jmespath

jsonpickle = lazy_import.lazy_module("jsonpickle")
import regex as re
from decorator import decorator

from .modules.internal.colors import blue, cyan, green, magenta, red, yellow

LOGGER = logging.getLogger("chepy")


class ChepyDecorators(object):
    """A class to house all the decorators for Chepy"""

    @staticmethod
    @decorator
    def call_stack(func, *args, **kwargs):
        """This decorator is used to get the method name and
        arguments and save it to self.stack. The data from
        self.stack is predominantly used to save recepies.
        """
        func_sig = dict()
        func_self = args[0]
        func_sig["function"] = func.__name__

        bound_args = inspect.signature(func).bind(*args, **kwargs)
        bound_args.apply_defaults()

        func_arguments = dict(bound_args.arguments)
        del func_arguments["self"]
        func_sig["args"] = func_arguments
        func_self._stack.append(func_sig)

        return func(*args, **kwargs)  # lgtm [py/call-to-non-callable]

    @staticmethod
    @decorator
    def is_stdout(func, *args, **kwargs):  # pragma: no cover
        """Detect if method is being called from the cli"""
        if sys.stdout.isatty():
            LOGGER.warning(f"{func.__name__} may not work as expected on the cli")
        return func(*args, **kwargs)


class ChepyCore(object):
    """The ChepyCore class for Chepy is primarily used as an interface
    for all the current modules/classes in Chepy, or for plugin development.
    The `ChepyCore` class is what provides the various attributes like **states**,
    **buffers**, etc and is required to use and extend Chepy.

    Args:
        *data (tuple): The core class takes arbitrary number of arguments as *args.

    Attributes:
        states (dict): Contains all the current states. Each arg passed to
            the ChepyCore class will be considered a state.
        buffers (dict): Contains all the current buffers if a buffer is saved.
        state (Any): The data in the current state. The state changes each time a
            Chepy method is called.

    Returns:
        Chepy: The Chepy object.
    """

    def __init__(self, *data):
        self.states = dict(list(enumerate(data)))
        #: Holder for the initial state
        self._initial_states = dict(list(enumerate(data)))
        #: Value of the initial state
        self._current_index = 0
        self.buffers = dict()
        #: Alias for `write_to_file`
        self.write = self.write_to_file
        #: Alias for `out`
        self.bake = self.out
        #: Alias for `web`
        self.cyberchef = self.web
        #: Alias for `load_file`
        self.read_file = self.load_file
        #: Holds all the methods that are called/chained and their args
        self._stack = list()
        #: Holds register values
        self._registers = dict()
        # logger
        self._log = LOGGER

    @property
    def recipe(self) -> List[Dict[str, Union[str, Dict[str, Any]]]]:
        """Returns the current recipe. This is a list of dictionaries
        that contains the method name and the arguments.

        Returns:
            List[Dict[str, Union[str, Dict[str, Any]]]]: The recipe.
        """
        return self._stack

    @property
    def state(self):
        return self.states[self._current_index]

    @state.setter
    def state(self, val):
        self.states[self._current_index] = val

    def __str__(self):
        try:
            if isinstance(self.state, bytearray):
                return re.sub(rb"[^\x00-\x7f]", b".", self.state).decode()
            else:
                return self._convert_to_str()
        except UnicodeDecodeError:  # pragma: no cover
            return "Could not convert to str, but the data exists in the states. Use o, output or out() to access the values"
        except:  # pragma: no cover
            self._log.exception(
                "\n\nCannot print current state. Either chain with "
                "another method, or use one of the output methods "
                "Example: .o, .out, or .state\n\n"
            )
            return ""

    def _pickle_class(self, obj: Any) -> Any:
        """This method takes another object as an argument and
        pickels that into a json object using jsonpickel. The
        return value is a dictionary

        Args:
            obj (Any): Any object

        Returns:
            Any: unpickeled JSON as a python object.
        """
        return json.loads(jsonpickle.encode(obj, unpicklable=True))

    def _load_as_file(self) -> object:
        """This method is used when a function or a method expects
        a file path to load a file. Instead of passing a file path,
        this method allows passing an io.BytesIO object instead.

        Returns:
            object: io.BytesIO object
        """
        return io.BytesIO(self._convert_to_bytes())

    def _abs_path(self, path: str):
        """Returns the absolute path by expanding home dir

        Args:
            path (str): Path to expand

        Returns:
            object: Path object
        """
        return Path(path).expanduser().absolute()

    def _info_logger(self, data: str) -> None:
        """Just a binding for logger.info

        Args:
            data (str): Message to log

        Returns:
            Chepy: The Chepy object.
        """
        self._log.info(blue(data))
        return None

    def _warning_logger(self, data: str) -> None:  # pragma: no cover
        """Just a binding for logger.warning

        Args:
            data (str): Message to log

        Returns:
            Chepy: The Chepy object.
        """
        self._log.warning(yellow(data))
        return None

    def _error_logger(self, data: str) -> None:  # pragma: no cover
        """Just a binding for logger.error

        Args:
            data (str): Message to log

        Returns:
            Chepy: The Chepy object.
        """
        self._log.error(red(data))
        return None

    def subsection(
        self,
        pattern: Union[str, bytes],
        methods: List[Tuple[Union[str, object], dict]],
        group: int = 0,
    ):
        """Run specified methods over a subsection of the state. This method will always treat the state
        as bytes.

        Args:
            pattern (Union[str, bytes]): Regex pattern to match against.
            methods (List[Tuple[Union[str, object], dict]]): Required. List of tuples. The first value of the
                tuple is the method name, the second value is a dictionary of arguments.
            group (int, optional): Matching group. Defaults to 0.

        Returns:
            _type_: _description_
        """
        if isinstance(pattern, str):
            pattern = pattern.encode()

        old_state = self._convert_to_bytes()
        new_state = b""
        start = 0
        for matched in re.compile(pattern).finditer(old_state):
            end, newstart = matched.span()
            self.state = matched.group(group)
            new_state += old_state[start:end]
            for method in methods:
                if type(method[0]).__name__ == "method":
                    method_name = method[0].__name__  # type: ignore
                elif isinstance(method[0], str):
                    method_name = method[0]
                if len(method) > 1:
                    getattr(self, method_name)(**method[1]).o
                else:
                    getattr(self, method_name)().o
            start = newstart
            new_state += self._convert_to_bytes()

        new_state += old_state[start:]

        self.state = new_state
        return self

    def fork(self, methods: List[Tuple[Union[str, object], dict]]):
        """Run multiple methods on all available states

        Method names in a list of tuples. If using in the cli,
        this should not contain any spaces.

        Args:
            methods (List[Tuple[Union[str, object], dict]]): Required. List of tuples

        Returns:
            Chepy: The Chepy object.

        Examples:
            This method takes an array of method names and their args as an list of
            tuples; the first value of the tuple is the method name as either a string,
            or as an object, and the second value is a ditionary of arguments. The keys of
            in the dictionary are method argument names, while the values are argument
            values.

            >>> from chepy import Chepy
            >>> c = Chepy("some", "data")
            >>> c.fork([("to_hex",), ("hmac_hash", {"secret_key": "key"})])
            >>> # this is how to use fork methods with a string
            >>> c.fork([(c.to_hex,), (c.hmac_hash, {"secret_key": "key"})])
            >>> # This is how to use fork using methods
            >>> print(c.states)
            {0: 'e46dfcf050c0a0d135b73856ab8e3298f9cc4105', 1: '1863d1542629590e3838543cbe3bf6a4f7c706ff'}
        """
        for i in self.states:
            self.change_state(i)
            for method in methods:
                if type(method[0]).__name__ == "method":
                    method_name = method[0].__name__  # type: ignore
                elif isinstance(method[0], str):
                    method_name = method[0]
                if len(method) > 1:
                    self.states[i] = getattr(self, method_name)(**method[1]).o
                else:
                    self.states[i] = getattr(self, method_name)().o
        return self

    def for_each(
        self,
        methods: List[Tuple[Union[str, object], dict]],
        merge: Union[str, bytes, None] = None,
    ):
        """Run multiple methods on current state if it is a list

        Method names in a list of tuples. If using in the cli,
        this should not contain any spaces.

        Args:
            methods (List[Tuple[Union[str, object], dict]]): Required.
                List of tuples
            merge (Union[str, bytes, None]): Merge data with. Defaults to None

        Returns:
            Chepy: The Chepy object.

        Examples:
            This method takes an array of method names and their args as an list of
            tuples; the first value of the tuple is the method name as either a string,
            or as an object, and the second value is a ditionary of arguments. The keys of
            in the dictionary are method argument names, while the values are argument
            values.

            >>> from chepy import Chepy
            >>> c = Chepy(['41', '42'])
            >>> c.for_each([("from_hex",), ("to_hex",)])
            >>> # this is how to use fork methods with a string
            >>> c.for_each([(c.from_hex,), (c.to_hex,)])
            >>> # This is how to use fork using methods
            >>> print(c)
            ['41', '42']
        """
        assert isinstance(self.state, list), "Current state is not a list"
        hold = self.state
        for i, val in enumerate(hold):
            self.state = val
            for method in methods:
                if type(method[0]).__name__ == "method":
                    method_name = method[0].__name__  # pragma: no cover
                elif isinstance(method[0], str):
                    method_name = method[0]
                if len(method) > 1:
                    hold[i] = getattr(self, method_name)(
                        **method[1]
                    ).o  # pragma: no cover
                else:
                    hold[i] = getattr(self, method_name)().o
        if merge is not None:
            if len(hold) > 0:
                merge = self._to_bytes(merge)
                if isinstance(hold[0], str):  # pragma: no cover
                    self.state = merge.decode().join(hold)
                elif isinstance(hold[0], bytes):
                    self.state = merge.join(hold)
                else:  # pragma: no cover
                    self.state = hold
        else:
            self.state = hold
        return self

    @ChepyDecorators.call_stack
    def set_state(self, data: Any):
        """Set any arbitrary values in the current state

        This method is simply changing the value of the instantiated
        state with an arbitrary value.

        Args:
            data (Any): Any data type

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("some data")
            >>> print(c.state)
            some data
            >>> c.set_state("New data")
            >>> print(c.state)
            New data
        """
        self.state = data
        return self

    @ChepyDecorators.call_stack
    def create_state(self):
        """Create a new empty state

        Returns:
            Chepy: The Chepy object.
        """
        self.states[len(self.states)] = {}
        return self

    @ChepyDecorators.call_stack
    def copy_state(self, index: int = None):
        """Copy the current state to a new state

        Args:
            index (int): Index of new state. Defaults to next available.

        Returns:
            Chepy: The Chepy object.
        """
        if not index:
            index = len(self.states)
        self.states[index] = self.states.get(self._current_index)
        return self

    @ChepyDecorators.call_stack
    def change_state(self, index: int):
        """Change current state by index

        Same behaviour as switch_state

        Args:
            index (int): Index of new state

        Raises:
            TypeError: If specified index does not exist

        Returns:
            Chepy: The Chepy object.
        """
        if index > len(self.states):  # pragma: no cover
            raise TypeError("Specified index does not exist")
        self._current_index = index
        return self

    @ChepyDecorators.call_stack
    def switch_state(self, index: int):  # pragma: no cover
        """Switch current state by index

        Same behaviour as change_state

        Args:
            index (int): Index of new state

        Raises:
            TypeError: If specified index does not exist

        Returns:
            Chepy: The Chepy object.
        """
        if index > len(self.states):  # pragma: no cover
            raise TypeError("Specified index does not exist")
        self._current_index = index
        return self

    @ChepyDecorators.call_stack
    def delete_state(self, index: int):
        """Delete a state specified by the index

        Args:
            index (int): Index of state

        Returns:
            Chepy: The Chepy object.
        """
        try:
            del self.states[index]
        except KeyError:  # pragma: no cover
            self._log.warning("{} does not exist".format(index))
        return self

    @ChepyDecorators.call_stack
    def get_state(self, index: int) -> Any:
        """Returns the value of the specified state.

        This method does not chain with other methods of Chepy

        Args:
            index (int): The index of the state

        Returns:
            Any: Any value that is in the specified state
        """
        return self.states.get(index)

    @ChepyDecorators.call_stack
    def save_buffer(self, index: int = None):
        """Save current state in a buffer

        Buffers are temporary holding areas for anything that is in the state.
        The state can change, but the buffer does not. Can be chained with other
        methods. Use in conjunction with `load_buffer` to load buffer back into
        the state.

        Args:
            index (int, optional): The index to save the state in, defaults to next index if None

        Returns:
            Chepy: The Chepy object.
        """
        if index is not None:
            self.buffers[index] = self.state
        else:
            self.buffers[len(self.buffers)] = self.state
        return self

    @ChepyDecorators.call_stack
    def load_buffer(self, index: int):
        """Load the specified buffer into state

        Args:
            index (int): Index key of an existing buffer

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("A").save_buffer()
            >>> # this saves the current value of state to a new buffer
            >>> c.to_hex()
            >>> # operate on a state, in this case, convert to hex.
            >>> c.state
            "41"
            >>> c.buffers
            {0: "A"}
            >>> c.load_buffer(0)
            >>> # loads the content of the buffer back into the current state.
            >>> c.state
            "A"
        """
        self.state = self.buffers[index]
        return self

    @ChepyDecorators.call_stack
    def delete_buffer(self, index: int):
        """Delete a buffer item

        Args:
            index (int): Key of buffer item

        Returns:
            Chepy: The Chepy object.
        """
        try:
            del self.buffers[index]
        except KeyError:  # pragma: no cover
            self._log.warning("{} does not exist".format(index))
        return self

    @ChepyDecorators.call_stack
    def substring(self, pattern: str, group: int = 0):
        """Choose a substring from current state as string

        The preceding methods will only run on the substring and
        not the original state. Group capture is supported.

        Args:
            pattern (str): Pattern to match.
            group (int, optional): Group to match. Defaults to 0.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.search(pattern, self._convert_to_str()).group(group)
        return self

    def _convert_to_bytes(self) -> bytes:
        """This method is used to coerce the current object in
        the state variable into a bytes. The method should be
        called inside any method that operates on a string object
        instead of calling `self.state` directly to avoid errors.

        Raises:
            NotImplementedError: If type coercian isn't available
                for the current state type.
        """
        if isinstance(self.state, bytes):
            return self.state
        elif isinstance(self.state, str):
            return self.state.encode()
        elif isinstance(self.state, int):
            return str(self.state).encode()
        elif isinstance(self.state, dict):
            return str(self.state).encode()
        elif isinstance(self.state, list):
            return str(self.state).encode()
        elif isinstance(self.state, bool):  # pragma: no cover
            return str(self.state).encode()
        elif isinstance(self.state, bytearray):
            return bytes(self.state)
        elif isinstance(self.state, float):
            return bytearray(struct.pack("f", self.state))
        else:  # pragma: no cover
            # todo check more types here
            raise NotImplementedError

    def _to_bytes(self, data: Any) -> bytes:  # pragma: no cover
        """This method is used to coerce data to bytes. The method should be
        called inside any method that operates on a string object
        instead of calling `self.state` directly to avoid errors.

        Raises:
            NotImplementedError: If type coercian isn't available
                for the current state type.
        """
        if isinstance(data, bytes):
            return data
        elif isinstance(data, str):
            return data.encode()
        elif isinstance(data, int):
            return str(data).encode()
        elif isinstance(data, dict):
            return str(data).encode()
        elif isinstance(data, list):
            return str(data).encode()
        elif isinstance(data, bool):
            return str(data).encode()
        elif isinstance(data, bytearray):
            return bytes(data)
        elif isinstance(data, float):
            return bytearray(struct.pack("f", data))
        else:  # pragma: no cover
            # todo check more types here
            raise NotImplementedError

    def _convert_to_bytearray(self) -> bytearray:
        """Attempts to coerce the current state into a
        `bytesarray` object
        """
        return bytearray(self._convert_to_bytes())

    def _convert_to_str(self) -> str:
        """This method is used to coerce the current object in
        the state variable into bytes. The method should be
        called inside any method that operates on a bytes object
        instead of calling `self.state` directly to avoid errors.

        Raises:
            NotImplementedError: If type coercian isn't available
                for the current state type.
        """
        if isinstance(self.state, bytes):
            return self.state.decode()
        elif isinstance(self.state, str):
            return self.state
        elif isinstance(self.state, int):
            return str(self.state)
        elif isinstance(self.state, dict):
            return str(self.state)
        elif isinstance(self.state, list):
            return str(self.state)
        elif isinstance(self.state, bool):  # pragma: no cover
            return str(self.state)
        elif isinstance(self.state, bytearray):
            return bytearray(self.state).decode()
        elif isinstance(self.state, float):  # pragma: no cover
            return format(self.state, "f")
        else:  # pragma: no cover
            # todo check more types here
            raise NotImplementedError

    def _str_to_bytes(self, s: str) -> bytes:  # pragma: no cover
        """Converts a str to bytes

        Args:
            s (str): String

        Returns:
            bytes: Bytes
        """
        if s is None:
            return s
        if isinstance(s, bytes):
            return s
        return s.encode()

    def _bytes_to_str(self, s: bytes) -> str:  # pragma: no cover
        """Converts a bytes to str

        Args:
            s (str): String

        Returns:
            str: String
        """
        if s is None:
            return s
        if isinstance(s, bytes):
            return s.decode()
        return s

    def _convert_to_int(self) -> int:
        """This method is used to coerce the current object in
        the state variable into an int. The method should be
        called inside any method that operates on a int types
        instead of calling `self.state` directly to avoid errors.

        Raises:
            NotImplementedError: If type coercian isn't available
                for the current state type.
        """
        if isinstance(self.state, int):
            return self.state
        elif isinstance(self.state, str) or isinstance(self.state, bytes):
            return int(self.state)
        else:  # pragma: no cover
            raise NotImplementedError

    @property
    def o(self):
        """Get the final output

        Returns:
            Any: Final output
        """
        if isinstance(self.state, str):
            return self.state.encode()
        return self.state

    @property
    def out_as_any(self):
        """Get the final output

        Returns:
            Any: Final output
        """
        return self.state

    @property
    def out(self) -> Any:
        """Get the final output

        Returns:
            Any: Final output
        """
        if isinstance(self.state, str):
            return self.state.encode()
        return self.state

    @ChepyDecorators.call_stack
    def get_by_index(self, *indexes: int):
        """Get an item by specifying an index. If only one index is specified, the obj is return else a new list is returned

        Args:
            *indexes (int): Index numbers to get.

        Returns:
            Chepy: The Chepy object.
        """
        if len(indexes) == 1:
            self.state = self.state[int(indexes[0])]
        else:
            self.state = [self.state[int(index)] for index in indexes]
        return self

    def _get_nested_value(self, data, key, split_by="."):
        """Get a dict value based on a string key with dot notation. Supports array indexing.
        If split_by is None or "", returns only the first key
        Args:
            data (dict): Data
            key (str): Dict key in a dot notation and array
            split_by (str, optional): Chars to split key by. Defaults to ".".
        """
        if not split_by:
            return data[key]
        try:
            keys = key.split(split_by)
            for key in keys:
                if "[" in key:
                    # Extract the key and index
                    key, index_str = key.split("[")
                    index_str = index_str.rstrip("]").strip()
                    if index_str == "*":
                        data = [data[key][i] for i in range(len(data[key]))]
                    else:
                        index = int(index_str)
                        data = data[key][index]
                else:
                    if isinstance(data, list):
                        data = [
                            data[i][key] for i in range(len(data)) if key in data[i]
                        ]
                    else:
                        data = data[key] if key in data else data
            return data
        except Exception as e:  # pragma: no cover
            self._error_logger(e)
            return data

    @ChepyDecorators.call_stack
    def get_by_key(self, *keys: str, py_style: bool = False, split_key: str = "."):
        """This method support json keys support.

        Args:
            keys (Tuple[Union[Hashable, None]]): Keys to extract.
            split_key (str, optional): Split nested keys. Defaults to "."
            nested (bool, optional): If the specified keys are nested. Supports array indexing. Defaults to True

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy({"a":{"b": "c"}}).get_by_key('a.b')
            >>> 'c'
        """
        assert isinstance(
            self.state,
            (
                dict,
                list,
            ),
        ), "State does not contain valid data"

        if py_style:
            if len(keys) == 1:
                self.state = self._get_nested_value(
                    self.state, keys[0], split_by=split_key
                )
            else:
                self.state = [
                    self._get_nested_value(self.state, key, split_by=split_key)
                    for key in keys
                ]
        else:
            o = jmespath.search(keys[0], self.state)
            if o is None:  # pragma: no cover
                raise ValueError("Query did not match any data")
            self.state = o
        return self

    @ChepyDecorators.call_stack
    def copy_to_clipboard(self) -> None:  # pragma: no cover
        """Copy to clipboard

        Copy the final output to the clipboard. If an
        error is raised, refer to the documentation on the error.

        Returns:
            None: Copies final output to the clipboard
        """
        pyperclip.copy(self._convert_to_str())
        return None

    @ChepyDecorators.call_stack
    def copy(self) -> None:  # pragma: no cover
        """Copy to clipboard

        Copy the final output to the clipboard. If an
        error is raised, refer to the documentation on the error.

        Returns:
            None: Copies final output to the clipboard
        """
        self.copy_to_clipboard()
        return None

    @ChepyDecorators.call_stack
    def web(
        self,
        magic: bool = False,
        cyberchef_url: str = "https://gchq.github.io/CyberChef/",
    ) -> None:  # pragma: no cover
        """Opens the current string in CyberChef on the browser as hex

        Args:
            magic (bool, optional): Start with the magic method in CyberChef
            cyberchef_url (string, optional): Base url for Cyberchef

        Returns:
            None: Opens the current data in CyberChef
        """
        data = re.sub(
            b"=+$", "", base64.b64encode(binascii.hexlify(self._convert_to_bytes()))
        )
        if magic:
            url = urljoin(
                cyberchef_url,
                "#recipe=From_Hex('None')Magic(3,false,false,'')&input={}".format(
                    data.decode()
                ),
            )
        else:
            url = urljoin(
                cyberchef_url,
                "#recipe=From_Hex('None')&input={}".format(data.decode()),
            )
        webbrowser.open_new_tab(url)
        return None

    @ChepyDecorators.call_stack
    def http_request(
        self,
        method: str = "GET",
        params: dict = {},
        json: dict = None,
        headers: dict = {},
        cookies: dict = {},
    ):  # pragma: no cover
        """Make a http/s request

        Make a HTTP/S request and work with the data in Chepy. Most common http
        methods are supported; but some methods may not provide a response body.

        Args:
            method (str, optional): Request method. Defaults to 'GET'.
            params (dict, optional): Query Args. Defaults to {}.
            json (dict, optional): Request payload. Defaults to None.
            headers (dict, optional): Headers for request. Defaults to {}.
            cookies (dict, optional): Cookies for request. Defaults to {}.

        Raises:
            NotImplementedError: If state is not a string or dictionary
            requests.RequestException: If response status code is not 200

        Returns:
            Chepy: A dictionary containing body, status and headers. The Chepy object.

        Examples:
            By default, this methed with make a GET request, But supports most
            common methods.

                >>> c = Chepy("http://example.com").http_request()
                >>> c.get_by_key("headers")

            This method can also be used to make more complex requests by specifying
            headers, cookies, body data etc.

                >>> c = Chepy("https://en4qpftrmznwq.x.pipedream.net")
                >>> c.http_request(
                >>>    method="POST",
                >>>    headers={"My-header": "some header"},
                >>>    json={"some": "data"}
                >>> )
                >>> print(c.get_by_key("body"))
                {"success": true}
        """

        def json2str(obj):  # pragma: no cover
            if isinstance(obj, dict):
                return obj
            elif isinstance(obj, str):
                return json.loads(obj)
            else:
                raise NotImplementedError

        try:
            from requests import request
        except ImportError:  # pragma: no cover
            self._error_logger("Could not import requests. pip install requests")
            return self

        params = json2str(params)
        headers = json2str(headers)
        cookies = json2str(cookies)
        res = request(
            method=method,
            url=self.state,
            params=params,
            json=json,
            headers=headers,
            cookies=cookies,
        )
        self.state = {
            "body": res.text,
            "status": res.status_code,
            "headers": dict(res.headers),
        }
        return self

    @ChepyDecorators.call_stack
    def load_from_url(
        self,
        method: str = "GET",
        params: dict = {},
        json: dict = None,
        headers: dict = {},
        cookies: dict = {},
    ):  # pragma: no cover
        """Load binary content from a url

        Most common http methods are supported; but some methods may not provide a response body.

        Args:
            method (str, optional): Request method. Defaults to 'GET'.
            params (dict, optional): Query Args. Defaults to {}.
            json (dict, optional): Request payload. Defaults to None.
            headers (dict, optional): Headers for request. Defaults to {}.
            cookies (dict, optional): Cookies for request. Defaults to {}.

        Raises:
            NotImplementedError: If state is not a string or dictionary
            requests.RequestException: If response status code is not 200

        Returns:
            Chepy: A bytearray of the response content. The Chepy object.

        Examples:
            By default, this methed with make a GET request, But supports most
            common methods.

                >>> c = Chepy("http://example.com/file.png").load_from_url()
                >>> b'\\x89PNG...'
        """

        def json2str(obj):  # pragma: no cover
            if isinstance(obj, dict):
                return obj
            elif isinstance(obj, str):
                return json.loads(obj)
            else:
                raise NotImplementedError

        try:
            from requests import request
        except ImportError:  # pragma: no cover
            self._error_logger("Could not import requests. pip install requests")
            return self

        params = json2str(params)
        headers = json2str(headers)
        cookies = json2str(cookies)
        res = request(
            method=method,
            url=self.state,
            params=params,
            json=json,
            headers=headers,
            cookies=cookies,
        )
        self.state = io.BytesIO(res.content).read()
        return self

    @ChepyDecorators.call_stack
    def load_dir(self, pattern: str = "*"):
        """Load all file paths in a directory

        Args:
            pattern (str, optional): File pattern to match. Defaults to "*".

        Returns:
            Chepy: The Chepy object.
        """
        files = [x for x in Path(self.state).glob(pattern) if x.is_file()]
        self.states = {x[0]: str(x[1]) for x in enumerate(files) if x[1].is_file()}
        return self

    @ChepyDecorators.call_stack
    def load_file(self, binary_mode: bool = False, encoding: Union[str, None] = None):
        """If a path is provided, load the file

        Args:
            binary_mode (bool, optional): Force load in binary mode.
            encoding (Union[str, None], optional): Encoding for string.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("/path/to/file")
            >>> # at the moment, the state only contains the string "/path/to/file"
            >>> c.load_file() # this will load the file content into the state
        """
        path = Path(str(self.state)).expanduser().absolute()
        if binary_mode:
            with open(path, "rb") as f:
                self.states[self._current_index] = bytearray(f.read())
        else:
            try:
                with open(path, "r", encoding=encoding) as f:  # type: ignore
                    self.states[self._current_index] = f.read()
            except UnicodeDecodeError:
                with open(path, "rb") as f:
                    self.states[self._current_index] = bytearray(f.read())
        return self

    def write_to_file(self, path: str) -> None:
        """Save the state to disk. Return None.

        Args:
            path (str): The file path to save in.

        Returns:
            None: Returns None

        Examples:
            >>> c = Chepy("some data").write_to_file('/some/path/file', as_binary=True)
        """
        if isinstance(path, bytes):  # pragma: no cover
            path = path.decode()
        with open(str(self._abs_path(path)), "w+") as f:
            f.write(self._convert_to_str())
        self._info_logger("File written to {}".format(self._abs_path(path)))
        return None

    def write_binary(self, path: str) -> None:  # pragma: no cover
        """Save the state to disk. Return None.

        Args:
            path (str): The file path to save in.

        Returns:
            None: Returns None

        Examples:
            >>> c = Chepy("some data").write_binary('/some/path/file')
        """
        data = self._convert_to_bytes()
        if isinstance(path, bytes):  # pragma: no cover
            path = path.decode()
        with open(str(self._abs_path(path)), "wb+") as f:
            f.write(data)
        self._info_logger("File written to {}".format(self._abs_path(path)))
        return None

    def run_recipe(self, recipes: List[Mapping[str, Union[str, Mapping[str, Any]]]]):
        """Run a recipe on the state. All arguments including optional needs to
        be specified for a recipe.

        Args:
            recipes (List[Mapping[str, Union[str, Mapping[str, Any]]]]): An array of recipes.
                Recipes are in the format {'function': 'function_name', 'args': {'arg_name': 'arg_val'}}

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy('bG9sCg==').run_recipe([{"function":"base64_decode","args":{"custom":None}}]])
            >>> lol
            In this example, we are calling the base64 decode method on the state.
        """
        for recipe in recipes:
            function = recipe["function"]
            args = recipe["args"]
            if len(args) > 0:
                getattr(self, function)(**args)
            else:
                getattr(self, function)()
        return self

    def save_recipe(self, path: str):
        """Save the current recipe

        A recipe will be all the previous methods called on the
        chepy instance along with their args

        Args:
            path (str): The path to save the recipe

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("some data").to_hex().base64_encode()
            >>> c.save_recipe("/path/to/recipe)
            >>> c.out
            NzM2ZjZkNjUyMDY0NjE3NDYx
        """
        with self._abs_path(path) as f:
            f.write_text(json.dumps(self._stack))
        self._info_logger("Saved recipe to {}".format(str(path)))
        return self

    def load_recipe(self, path: str):
        """Load and run a recipe

        Args:
            path (str): Path to recipe file

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("some data").load_recipe("/path/to/recipe").out
            NzM2ZjZkNjUyMDY0NjE3NDYx
        """
        with self._abs_path(path) as f:
            recipes = json.loads(f.read_text())
            for recipe in recipes:
                function = recipe["function"]
                args = recipe["args"]
                if len(args) > 0:
                    getattr(self, function)(**args)
                else:
                    getattr(self, function)()
        return self

    # @ChepyDecorators.call_stack
    def run_script(self, path: str, save_state: bool = False):
        """Inject and run a custom script on the state.
        The custom script must have a function called **cpy_script** which
        must take one argument. The state is passed as the argument.

        Args:
            path (str): Path to custom script
            save_state (bool, optional): Save script output to the state. Defaults to False.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("A").to_hex().run_script('tests/files/script.py', True)
            b'4141'
        """
        script_path = str(self._abs_path(path))
        loader = SourceFileLoader("cpy_s", script_path)
        handle = loader.load_module("cpy_s")
        if save_state:
            self.state = handle.cpy_script(self.state)
        else:
            print(cyan("Script Output: {}".format(script_path)))
            print(handle.cpy_script(self.state))
        return self

    @ChepyDecorators.call_stack
    def loop(self, iterations: int, callback: str, args: dict = {}):
        """Loop and apply callback n times

        Args:
            iterations (int): Number of iterations to loop
            callback (str): The Chepy method to loop over
            args (dict, optional): Optional arguments for the callback. Defaults to {}.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("VmpGb2QxTXhXWGxTYmxKV1lrZDRWVmx0ZEV0alZsSllaVWRHYWxWVU1Eaz0=")
            >>> c.loop(iterations=6, callback='hmac_hash', args={'key': 'secret'})
            securisec
        """
        if type(callback).__name__ == "method":
            # this allows for both method and string passing
            callback = callback.__name__
        assert isinstance(callback, str), "Callback must be a string"
        assert isinstance(iterations, int), "Iterations must be an integer"
        assert isinstance(args, dict), "Args must be a dick"

        stack_loop_index = next(
            itertools.dropwhile(
                lambda x: self._stack[x]["function"] != "loop",
                reversed(range(len(self._stack))),
            )
        )

        for _ in range(int(iterations)):
            getattr(self, callback)(**args)

        self._stack = self._stack[: stack_loop_index + 1]
        return self

    @ChepyDecorators.call_stack
    def loop_list(self, callback: str, args: dict = {}):
        """Loop over an array and run a Chepy method on it

        Args:
            callback (str): Chepy method as string
            args (dict, optional): Dictionary of args. If in cli, dont use spaces. Defaults to {}.

        Returns:
            Chepy: The Chepy object

        Examples:
            This method is capable of running a callable from either
            a string, or a chepy method.

            >>> c = Chepy(["an", "array"])
            >>> c.loop_list('to_hex').loop_list('hmac_hash', {'key': 'secret'})
            ['5cbe6ca2a66b380aec1449d4ebb0d40ac5e1b92e', '30d75bf34740e8781cd4ec7b122e3efd8448e270']
        """
        if type(callback).__name__ == "method":
            # this allows for both method and string passing
            callback = callback.__name__

        assert isinstance(self.state, list), "State is not a list"
        assert isinstance(callback, str), "Callback must be a string"
        hold = []
        current_state = self.state
        # find the last index that this method was run
        stack_loop_index = next(
            itertools.dropwhile(
                lambda x: self._stack[x]["function"] != "loop_list",
                reversed(range(len(self._stack))),
            )
        )
        if isinstance(args, str):  # pragma: no cover
            args = json.loads(args)
        try:
            for index, data in enumerate(current_state):
                self.state = current_state[index]
                if args:
                    hold.append(getattr(self, callback)(**args).o)
                else:
                    hold.append(getattr(self, callback)().o)
            self._stack = self._stack[: stack_loop_index + 1]
            self.state = hold
            return self
        except:  # pragma: no cover
            self.state = current_state
            raise

    @ChepyDecorators.call_stack
    def loop_dict(self, keys: list, callback: str, args: dict = {}):
        """
        Loop over a dictionary and apply the callback to the value

        Args:
            keys (list): List of keys to match. If in cli, dont use spaces.
            callback (str): Chepy method as string
            args (dict, optional): Dictionary of args. If in cli, dont use spaces. Defaults to {}.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy({'some': 'hahahaha', 'lol': 'aahahah'})
            >>> c.loop_dict(['some'], 'hmac_hash', {'key': 'secret'}).o
            {'some': '99f77ec06a3c69a4a95371a7888245ba57f47f55', 'lol': 'aahahah'}

            We can combine `loop_list` and `loop_dict` to loop over a list of dictionaries.

            >>> data = [{"some": "val"}, {"some": "another"}, {"lol": "lol"}, {"another": "aaaa"}]
            >>> c = Chepy(data)
            >>> c.loop_list("loop_dict", {"keys": ["some", "lol"], "callback": "to_upper_case"})
            [
                {"some": "VAL"},
                {"some": "ANOTHER"},
                {"lol": "LOL"},
                {"another": "aaaa"},
            ]
        """
        if type(callback).__name__ == "method":
            # this allows for both method and string passing
            callback = callback.__name__
        assert isinstance(callback, str), "Callback must be a string"

        hold = {}
        current_state = self.state
        # find the last index that this method was run
        stack_loop_index = next(
            itertools.dropwhile(
                lambda x: self._stack[x]["function"] != "loop_dict",
                reversed(range(len(self._stack))),
            )
        )

        if isinstance(keys, str):  # pragma: no cover
            keys = json.loads(keys)

        if isinstance(args, str):  # pragma: no cover
            args = json.loads(args)
        try:
            dict_keys = current_state.keys()
            for key in keys:
                if current_state.get(key) is not None:
                    self.state = current_state.get(key)
                    if args:
                        hold[key] = getattr(self, callback)(**args).o
                    else:
                        hold[key] = getattr(self, callback)().o
            for unmatched_key in list(set(dict_keys) - set(keys)):
                hold[unmatched_key] = current_state[unmatched_key]
            self._stack = self._stack[: stack_loop_index + 1]
            self.state = hold
            return self
        except:  # pragma: no cover
            self.state = current_state
            raise

    @ChepyDecorators.call_stack
    def debug(self, verbose: bool = False):
        """Debug the current instance of Chepy

        This method does not change the state.

        Args:
            verbose (bool, optional): Show verbose info. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        print(cyan("Current state:"), yellow(str(self._current_index)))
        print(cyan("Current states:"), yellow(str(len(self.states))))
        print(
            cyan("Current state types:"),
            yellow(str({k: type(v).__name__ for k, v in self.states.items()})),
        )
        print(cyan("Current buffers:"), yellow(str(len(self.buffers))))
        print(
            cyan("Current buffer types:"),
            yellow(str({k: type(v).__name__ for k, v in self.buffers.items()})),
        )
        if verbose:
            print(magenta("States:"), self.states)
            print(magenta("Buffers:"), self.buffers)
        return self

    @ChepyDecorators.call_stack
    def print(self, *args):  # pragma: no cover
        """Print the state

        Returns:
            Chepy: The Chepy object.
        """
        print(*args, self.state)
        return self

    @ChepyDecorators.call_stack
    def reset(self):
        """Reset states back to their initial values

        Returns:
            Chepy: The Chepy object.
        """
        self.states = self._initial_states
        return self

    @ChepyDecorators.call_stack
    def load_command(self):  # pragma: no cover
        """Run the command in state and get the output

        Returns:
            Chepy: The Chepy object.

        Examples:
            This method can be used to interface with the shell and Chepy
            directly by ingesting a commands output in Chepy.

            >>> c = Chepy("ls -l").shell_output().o
            test.html
            ...
            test.py
        """
        self.state = subprocess.getoutput(self.state)
        return self

    @ChepyDecorators.call_stack
    def pretty(self, indent: int = 2):  # pragma: no cover
        """Prettify the state.

        Args:
            indent (int, optional): Indent level. Defaults to 2.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = pformat(self.state, indent=int(indent))
        return self

    def plugins(self, enable: str) -> None:  # pragma: no cover
        """Use this method to enable or disable Chepy plugins.

        Valid options are `true` or `false`. Once this method completes,
        it does call sys.exit().

        Args:
            enable (str): Set to `true` or `false`

        Returns:
            None
        """
        assert enable in ["true", "false"], "Valid values are true and false"
        conf_path = Path().home() / ".chepy" / "chepy.conf"
        c = ConfigParser()
        c.read(conf_path)
        c.set("Plugins", "enableplugins", enable)
        with open(conf_path, "w") as f:
            c.write(f)
        if enable:
            self._info_logger(
                green(
                    "Plugins have been enabled. Restart Chepy for effects to take place."
                )
            )
        else:
            self._info_logger(
                green(
                    "Plugins have been disabled. Restart Chepy for effects to take place."
                )
            )
        sys.exit()
        return None

    def set_plugin_path(self, path: str) -> None:  # pragma: no cover
        """Use this method to set the path for Chepy plugins.

        Args:
            path (str): Path to plugins directory

        Returns:
            None
        """
        expand_path = self._abs_path(path)
        if expand_path.exists():
            conf_path = Path().home() / ".chepy" / "chepy.conf"
            c = ConfigParser()
            c.read(conf_path)
            c.set("Plugins", "pluginpath", str(expand_path))
            with open(conf_path, "w") as f:
                c.write(f)
            self._info_logger(green("Plugin path has been set. Restart for changes."))
            sys.exit()
            return None
        else:
            raise AttributeError("The path does not exist")

    def callback(self, callback_function: Callable[[Any], Any]):
        """Run any user defined python function against the state.
        This method is not recorded in the recipes

        Args:
            callback_function (Callable[[Any], Any]): The function to run. The function should take one argument (the state is passed to it). It can return Any

        Examples:
            from chepy import Chepy

            def cb(data):
                return data * 2

            c = Chepy('abc').callback(cb)
            # state is now abcabc

        Returns:
            Chepy: The Chepy object.
        """
        # dont run if from cli
        if sys.stdout.isatty():  # pragma: no cover
            self._log.warning("callback cannot be used via the cli")
            return self
        self.state = callback_function(self.state)
        return self

    @ChepyDecorators.call_stack
    def register(
        self,
        pattern: Union[str, bytes],
        ignore_case: bool = False,
        multiline: bool = False,
        dotall: bool = False,
        unicode: bool = False,
        extended: bool = False,
    ):
        """Extract data from the input and store it in registers. Regular expression capture groups are used to select the data to extract.

        Args:
            pattern (Union[str, bytes]): Required. The regex pattern to search by
            ignore_case (bool, optional): Set case insensitive flag. Defaults to False.
            multiline (bool, optional): ^/$ match start/end. Defaults to False.
            dotall (bool, optional): `.` matches newline. Defaults to False.
            unicode (bool, optional): Match unicode characters. Defaults to False.
            extended (bool, optional): Ignore whitespace. Defaults to False.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("hello world")
            >>> c.register("(hello)\\s(world)")
            >>> c._registers
            {'$R0': 'hello', '$R1': 'world'}
        """
        # regex flags
        flags = 0
        if ignore_case:
            flags += re.IGNORECASE
        if multiline:
            flags += re.MULTILINE
        if dotall:
            flags += re.DOTALL
        if unicode:
            flags += re.UNICODE
        if extended:  # pragma: no cover
            flags += re.X

        r = re.compile(pattern, flags=flags)
        old_state = self.state

        if isinstance(pattern, bytes):
            matches = r.findall(self._convert_to_bytes())
        else:
            matches = r.findall(self._convert_to_str())

        # there are matches which is a list of tuples
        if len(matches) > 0:
            # if there is only one match, it will be a list. else, it is a list with a tuple
            if isinstance(matches[0], tuple):  # pragma: no cover
                matches = matches[0]
            for i in range(len(matches)):
                # only add non empty
                if matches[i]:
                    self._registers[f"$R{i}"] = matches[i]

        self.state = old_state
        return self

    @ChepyDecorators.call_stack
    def prefix(self, data: bytes):
        """Add a prefix to the data in state. The state is converted to bytes

        Args:
            data (bytes): Data to add

        Returns:
            Chepy: The Chepy object.
        """
        data = self._str_to_bytes(data)
        self.state = data + self._convert_to_bytes()
        return self

    @ChepyDecorators.call_stack
    def suffix(self, data: Union[str, bytes]):
        """Add a suffix to the data in state. The state is converted to bytes

        Args:
            data (bytes): Data to add

        Returns:
            Chepy: The Chepy object.
        """
        data = self._str_to_bytes(data)
        self.state = self._convert_to_bytes() + data
        return self

    def get_register(self, key: str) -> Union[str, bytes]:
        """Get a value from registers by key

        Args:
            key (str): Key

        Raises:
            ValueError: If key does not exist

        Returns:
            Union[str, bytes]: Value of register
        """
        v = self._registers.get(key)
        if v is None:  # pragma: no cover
            raise ValueError("Key not found in registers")
        return v

    def set_register(self, key: str, val: Union[str, bytes]):
        """Set the value of a register

        Args:
            key (str): Key
            val (Union[str, bytes]): Value

        Returns:
            Chepy: The Chepy object.
        """
        self._registers[key] = val
        return self

    @ChepyDecorators.call_stack
    def dump_json(self):
        """Json serialize the state

        Returns:
            Chepy: The Chepy object.
        """

        # Function to recursively convert bytes to UTF-8 strings or Base64-encoded strings
        def encode_bytes(obj):
            if isinstance(obj, bytes):
                try:
                    # Try to decode as UTF-8
                    return obj.decode("utf-8")
                except UnicodeDecodeError:  # pragma: no cover
                    # If decoding fails, encode as Base64
                    return base64.b64encode(obj).decode("utf-8")
            elif isinstance(obj, dict):
                return {
                    encode_bytes(k) if isinstance(k, bytes) else k: encode_bytes(v)
                    for k, v in obj.items()
                }
            elif isinstance(obj, list):
                return [encode_bytes(item) for item in obj]
            else:
                return obj

        self.state = json.dumps(encode_bytes(self.state))
        return self

    @ChepyDecorators.call_stack
    def walk_dir(self):
        """Walk a directory and get all file paths

        Returns:
            Chepy: The Chepy object.
        """
        p = self._convert_to_str()
        paths = Path(p).glob("**/*")
        self.state = [str(p) for p in paths]
        return self

    @ChepyDecorators.call_stack
    def search_dir(self, pattern: Union[bytes, str]):
        """Search all files in a directory. Pattern is case insensitive

        Args:
            pattern (Union[bytes, str]): regex to search

        Returns:
            Chepy: The Chepy object.
        """
        paths = self._convert_to_str()
        rgx = re.compile(self._str_to_bytes(pattern), flags=re.I)
        hold = []
        for path in Path(paths).glob("**/*"):
            if path.is_dir():  # pragma: no cover
                continue
            data = path.read_bytes()
            matched = rgx.findall(data)
            if matched:
                hold += matched
        self.state = hold
        return self



import base64
import binascii
import hmac
import hashlib
import itertools
import string
import random
import regex as re
import json
from typing import Literal, TypeVar, Dict, Any, Union
from .internal.ls47 import (
    encrypt_pad as _ls47_enc,
    decrypt_pad as _ls47_dec,
    derive_key as _derive_key,
)
from .internal.constants import Ciphers, Rabbit
from .internal.helpers import detect_delimiter
from .internal.helpers import Zeckendorf

import lazy_import

pgpy = lazy_import.lazy_module("pgpy")


AES = lazy_import.lazy_module("Crypto.Cipher.AES")
Salsa20 = lazy_import.lazy_module("Crypto.Cipher.Salsa20")
ARC4 = lazy_import.lazy_module("Crypto.Cipher.ARC4")
DES = lazy_import.lazy_module("Crypto.Cipher.DES")
ChaCha20 = lazy_import.lazy_module("Crypto.Cipher.ChaCha20")
DES3 = lazy_import.lazy_module("Crypto.Cipher.DES3")
RSA = lazy_import.lazy_module("Crypto.PublicKey.RSA")
Hash = lazy_import.lazy_module("Crypto.Hash")
Counter = lazy_import.lazy_module("Crypto.Util.Counter")
PKCS1_15 = lazy_import.lazy_module("Crypto.Signature.pkcs1_15")
PKCS1_OAEP = lazy_import.lazy_module("Crypto.Cipher.PKCS1_OAEP")
PKCS1_v1_5 = lazy_import.lazy_module("Crypto.Cipher.PKCS1_v1_5")
Blowfish = lazy_import.lazy_module("Crypto.Cipher.Blowfish")
Padding = lazy_import.lazy_module("Crypto.Util.Padding")
pycipher = lazy_import.lazy_module("pycipher")
Fernet = lazy_import.lazy_callable("cryptography.fernet.Fernet")

from ..core import ChepyCore, ChepyDecorators
from ..extras.combinatons import hex_chars
from .internal.constants import EncryptionConsts

EncryptionEncodingT = TypeVar("EncryptionEncodingT", bound="EncryptionEncoding")


class EncryptionEncoding(ChepyCore):
    """This class handles most operations related to various encryption
    related operations. This class inherits the ChepyCore class, and all the
    methods are also available from the Chepy class

    Examples:
        >>> from chepy import Chepy
        or
        >>> from chepy.modules.encryptionencoding import EncryptionEncoding
    """

    def __init__(self, *data):
        super().__init__(*data)

    def __check_mode(self, mode) -> None:
        assert mode in ["CBC", "OFB", "CTR", "ECB"], "Not a valid mode."

    def _convert_key(
        self, key, iv, key_format: str, iv_format: str
    ) -> EncryptionEncodingT:  # pragma: no cover
        key = self._str_to_bytes(key)
        # modify key according to mode
        if key_format == "hex":
            key = binascii.unhexlify(key)
        if key_format == "base64" or key_format == "b64":
            key = base64.b64decode(key)
        if key_format == "utf-8" or key_format == "utf8":
            key = key.decode().encode("utf-8")
        if key_format == "latin-1":
            key = key.decode().encode("latin-1")
        if key_format == "raw":
            key = key

        # modify iv according to mode
        iv = self._str_to_bytes(iv)
        if iv_format == "hex":
            iv = binascii.unhexlify(iv)
        if iv_format == "base64" or iv_format == "b64":
            iv = base64.b64decode(iv)
        if iv_format == "utf-8" or iv_format == "utf8":
            iv = iv.decode().encode("utf-8")
        if iv_format == "latin-1":
            iv = iv.decode().encode("latin-1")
        if iv_format == "raw":
            iv = iv
        else:
            iv = binascii.unhexlify(binascii.hexlify(iv))
        return key, iv

    def _rsa_process_key(self, key: str, is_file: bool, passphrase=None):
        """Returns an RSA instance for either keyfile or string key"""
        if is_file:
            with open(str(self._abs_path(key)), "r") as f:
                return RSA.import_key(f.read(), passphrase)
        return RSA.import_key(key, passphrase)

    def _rsa_cipher(self, c_format: str, rsa):
        """Returns an RSA Cipher instance based for cipher type"""
        if c_format == "PKCS":
            return PKCS1_v1_5.new(rsa)
        return PKCS1_OAEP.new(rsa)

    @ChepyDecorators.call_stack
    def rotate(self, rotate_by: int) -> EncryptionEncodingT:
        """Rotate string by provided number

        Args:
            rotate_by (int): Required. Number to rotate by

        Returns:
            Chepy: The Chepy object.

        Examples:
            In this example, we will rotate by 20

            >>> Chepy("some data").rotate(20).out
            "migy xunu"
        """
        lc = string.ascii_lowercase
        uc = string.ascii_uppercase
        lookup = str.maketrans(
            lc + uc, lc[rotate_by:] + lc[:rotate_by] + uc[rotate_by:] + uc[:rotate_by]
        )
        self.state = self.state.translate(lookup).encode()
        return self

    @ChepyDecorators.call_stack
    def rotate_bruteforce(self) -> EncryptionEncodingT:
        """Brute force rotation from 1 to 26.
        Returned value is a dict where key is the rotation count.

        Returns:
            Chepy: The Chepy object.

        Examples:
            In this example, we will rotate by 20

            >>> Chepy('uryyb').rotate_bruteforce()
            {
                '1': 'vszzc',
                '2': 'wtaad',
                ...
                '13': 'hello',
                ...
            }
        """
        hold = {}
        lc = string.ascii_lowercase
        uc = string.ascii_uppercase
        for rotate_by in range(1, 27):
            lookup = str.maketrans(
                lc + uc,
                lc[rotate_by:] + lc[:rotate_by] + uc[rotate_by:] + uc[:rotate_by],
            )
            hold[str(rotate_by)] = self.state.translate(lookup).encode()
        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def rot_13(
        self, amount=13, rotate_lower=True, rotate_upper=True, rotate_numbers=False
    ) -> EncryptionEncodingT:
        """Rot 13

        Args:
            amount (int, optional): Rotate amount. Defaults to 13.
            rotate_lower (bool, optional): Rotate lowercase. Defaults to True.
            rotate_upper (bool, optional): Rotate uppercase. Defaults to True.
            rotate_numbers (bool, optional): Rotate numbers. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        text = self._convert_to_str()
        result = []
        for char in text:
            if rotate_lower and "a" <= char <= "z":  # Lowercase letters
                result.append(chr((ord(char) - ord("a") + amount) % 26 + ord("a")))
            elif rotate_upper and "A" <= char <= "Z":  # Uppercase letters
                result.append(chr((ord(char) - ord("A") + amount) % 26 + ord("A")))
            elif rotate_numbers and "0" <= char <= "9":  # Numbers
                result.append(chr((ord(char) - ord("0") + amount) % 10 + ord("0")))
            else:
                result.append(char)  # Non-alphabetical characters remain unchanged
        self.state = "".join(result)
        return self

    @ChepyDecorators.call_stack
    def rot_47(self, rotation: int = 47) -> EncryptionEncodingT:
        """ROT 47 encoding

        A slightly more complex variation of a caesar cipher, which includes
        ASCII characters from 33 '!' to 126 '~'. Default rotation: 47.

        Args:
            rotation (int, optional): Amount to rotate by. Defaults to 14.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some").rot_47().out
            b"D@>6"
        """
        decoded_string = ""
        for char in self._convert_to_str():
            if ord(char) >= 33 and ord(char) <= 126:
                decoded_char = chr((ord(char) - 33 + rotation) % 94 + 33)
                decoded_string += decoded_char
            else:
                decoded_string += char  # pragma: no cover
        self.state = decoded_string.encode()
        return self

    @ChepyDecorators.call_stack
    def rot_47_bruteforce(self) -> EncryptionEncodingT:
        """ROT 47 bruteforce

        Returns:
            Chepy: The Chepy object.
        """
        hold = {}
        data = self._convert_to_str()
        for r in range(1, 94):
            decoded_string = ""
            for char in data:
                if ord(char) >= 33 and ord(char) <= 126:
                    decoded_char = chr((ord(char) - 33 + r) % 94 + 33)
                    decoded_string += decoded_char
                else:
                    decoded_string += char  # pragma: no cover
            hold[str(r)] = decoded_string.encode()
        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def rot_8000(self):
        """Rot8000

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_str()
        valid_code_points = {
            33: True,
            127: False,
            161: True,
            5760: False,
            5761: True,
            8192: False,
            8203: True,
            8232: False,
            8234: True,
            8239: False,
            8240: True,
            8287: False,
            8288: True,
            12288: False,
            12289: True,
            55296: False,
            57344: True,
        }

        BMP_SIZE = 0x10000

        rotlist = {}  # the mapping of char to rotated char
        hiddenblocks = []
        startblock = 0

        for key, value in valid_code_points.items():
            if value:
                hiddenblocks.append({"start": startblock, "end": key - 1})
            else:
                startblock = key

        validintlist = []  # list of all valid chars
        currvalid = False

        for i in range(BMP_SIZE):
            if i in valid_code_points:
                currvalid = valid_code_points[i]
            if currvalid:
                validintlist.append(i)

        rotatenum = len(validintlist) // 2

        # go through every valid char and find its match
        for i in range(len(validintlist)):
            rotlist[chr(validintlist[i])] = chr(
                validintlist[(i + rotatenum) % (rotatenum * 2)]
            )

        outstring = ""

        for char in data:
            # if it is not in the mappings list, just add it directly (no rotation)
            if char not in rotlist:
                outstring += char  # pragma: no cover
                continue  # pragma: no cover

            # otherwise, rotate it and add it to the string
            outstring += rotlist[char]

        self.state = outstring.encode()
        return self

    @ChepyDecorators.call_stack
    def xor(
        self,
        key: str,
        key_type: Literal["hex", "utf8", "base64", "decimal", "raw"] = "hex",
    ) -> EncryptionEncodingT:
        """XOR state with a key. In raw format, both the state and the key is converted
        to a bytearray.

        Args:
            key (str): Required. The key to xor by
            key_type (str, optional): The key type. Valid values are hex, utf8, decimal and base64. Defaults to "hex".

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("secret").xor(key="secret", key_type="utf8").to_hex()
            000000000000
        """

        x = bytearray(b"")

        if key_type == "raw":
            _data_ba = bytearray()
            _data = self._convert_to_str()
            _data_ba.extend(map(ord, _data))
            raw_key = bytearray()
            raw_key.extend(map(ord, key))
            for char, key_val in zip(_data_ba, itertools.cycle(raw_key)):
                x.append(char ^ key_val)

        # check if state is a list and keys are list
        elif isinstance(self.state, bytearray) and isinstance(key, bytearray):
            for char, key_val in zip(self.state, itertools.cycle(key)):
                x.append(char ^ key_val)

        else:
            if key_type == "utf8":
                key = str(key)
                key = binascii.hexlify(key.encode())
            elif key_type == "base64":
                key = binascii.hexlify(base64.b64decode(key.encode()))
            elif key_type == "decimal":
                if int(key) < 0 or int(key) > 255:
                    raise ValueError("Invalid decimal key")  # pragma: no cover

            _s = self._convert_to_bytes()

            if key_type == "decimal":
                for b in _s:
                    x.append(b ^ int(key))
            else:
                key = binascii.unhexlify(key)
                for char, key_val in zip(_s, itertools.cycle(key)):
                    x.append(char ^ key_val)

        self.state = bytes(x)
        return self

    @ChepyDecorators.call_stack
    def xor_bruteforce(
        self, length: int = 100, crib: Union[bytes, str, None] = None
    ) -> EncryptionEncodingT:
        """Brute force single byte xor

        For multibyte xor bruteforce, use chepy.extras.crypto_extras.xor_bruteforce_multi
        function

        Args:
            length (int, optional): How to bytes to bruteforce. Defaults to 100.
            crib (Union[bytes, str, None], optional): Check for crib in xored value. Defaults to None.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("pf`qfw").xor_bruteforce()
            {'00': bytearray(b'pf`qfw'),
            '01': bytearray(b'qgapgv'),
            '02': bytearray(b'rdbsdu'),
            '03': bytearray(b'secret'), # here is our secret xored with the hex key 03
            '04': bytearray(b'tbdubs'),
            '05': bytearray(b'ucetcr'),
            ...}
            >>> c.get_by_key("03").bytearray_to_str()
            secret
            >>> c.xor("03").bytearray_to_str()
            pf`qfw
        """
        crib = self._str_to_bytes(crib)
        original = self.state
        found = {}
        keys = hex_chars()
        self.state = original[:length]
        for key in keys:
            self.xor(key)
            if crib is not None:
                if crib in self.state:
                    found[key] = self.state
            else:
                found[key] = self.state
            self.state = original[:length]
        self.state = found
        return self

    @ChepyDecorators.call_stack
    def jwt_decode(self) -> EncryptionEncodingT:
        """Decode a JWT token. Does not verify

        Returns:
            Chepy: The Chepy object.
        """
        def _b64decode_pad(s):
            s += "=" * (-len(s) % 4)
            return json.loads(base64.urlsafe_b64decode(s))

        token = self._convert_to_str()
        parts = token.split(".")
        self.state = {
            "header": _b64decode_pad(parts[0]),
            "payload": _b64decode_pad(parts[1]),
        }
        return self

    @ChepyDecorators.call_stack
    def jwt_verify(
        self, secret: str, algorithm: list = ["HS256"]
    ) -> EncryptionEncodingT:
        """Verify JWT token

        Args:
            secret (str): Required. Secret key for token
            algorithm (list, optional): Array of valid algorithms. Defaults to ["HS256"]

        Returns:
            Chepy: The Chepy object.
        """
        token = self._convert_to_str().strip().replace(" ", "")
        parts = token.split(".")
        assert len(parts) == 3, "Invalid JWT format"

        alg = algorithm[0] if isinstance(algorithm, list) else algorithm
        hash_map = {"HS256": hashlib.sha256, "HS384": hashlib.sha384, "HS512": hashlib.sha512}
        assert alg in hash_map, f"Unsupported algorithm: {alg}"

        signing_input = f"{parts[0]}.{parts[1]}".encode()
        expected_sig = base64.urlsafe_b64encode(
            hmac.new(secret.encode(), signing_input, hash_map[alg]).digest()
        ).rstrip(b"=")
        actual_sig = parts[2].encode()
        assert hmac.compare_digest(expected_sig, actual_sig), "Signature verification failed"

        padded = parts[1] + "=" * (-len(parts[1]) % 4)
        self.state = json.loads(base64.urlsafe_b64decode(padded))
        return self

    @ChepyDecorators.call_stack
    def jwt_sign(self, secret: str, algorithms: str = "HS256") -> EncryptionEncodingT:
        """Sign a json/dict object in JWT

        Args:
            secret (str): Required. Secret to sign with
            algorithms (str, optional): Signing algorithm. Defaults to "HS256".

        Returns:
            Chepy: The Chepy object.
        """
        if isinstance(self.state, dict):
            data = self.state
        elif isinstance(self.state, str):
            data = json.loads(self.state)

        hash_map = {"HS256": hashlib.sha256, "HS384": hashlib.sha384, "HS512": hashlib.sha512}
        assert algorithms in hash_map, f"Unsupported algorithm: {algorithms}"

        header = base64.urlsafe_b64encode(json.dumps({"typ": "JWT", "alg": algorithms}, separators=(",", ":")).encode()).rstrip(b"=").decode()
        payload = base64.urlsafe_b64encode(json.dumps(data, separators=(",", ":")).encode()).rstrip(b"=").decode()
        signing_input = f"{header}.{payload}".encode()
        sig = base64.urlsafe_b64encode(
            hmac.new(secret.encode(), signing_input, hash_map[algorithms]).digest()
        ).rstrip(b"=").decode()
        self.state = f"{header}.{payload}.{sig}"
        return self

    @ChepyDecorators.call_stack
    def jwt_token_generate_none_alg(
        self, headers: Dict[str, Any] = {}
    ) -> EncryptionEncodingT:
        """Generate a jwt token with none algorithm

        Args:
            headers (Dict[str, Any], optional): Headers. `alg` key will be overwritten. Defaults to {}.

        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, dict), "State should be a dictionary"
        headers["alg"] = "none"
        encoded_headers = base64.b64encode(json.dumps(headers).encode()).replace(
            b"=", b""
        )
        encoded_payload = base64.b64encode(json.dumps(self.state).encode()).replace(
            b"=", b""
        )
        self.state = encoded_headers + b"." + encoded_payload + b"."
        return self

    @ChepyDecorators.call_stack
    def jwt_token_generate_embedded_jwk(
        self,
        private_key_pem: str,
        private_key_passphrase: str = None,
        headers: dict = {},
        alg: str = "RS256",
    ) -> EncryptionEncodingT:
        """Generate a JWT token with an embedded JWK

        Args:
            private_key_pem (str): Private key to sign token
            private_key_passphrase (str, optional): Private key passphrase. Defaults to None.
            headers (dict, optional): Token headers. Defaults to {}.
            alg (str, optional): Token algorithm. Defaults to "RS256".

        Returns:
            Chepy: The Chepy object.
        """
        payload = self.state
        assert isinstance(payload, dict), "State should be a dictionary"
        private_key = RSA.import_key(private_key_pem, private_key_passphrase)

        n = private_key.n
        e = private_key.e

        jwk_header = {
            "kty": "RSA",
            "e": base64.urlsafe_b64encode(e.to_bytes((e.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("="),
            "n": base64.urlsafe_b64encode(n.to_bytes((n.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("="),
        }
        if headers.get("kid"):
            jwk_header["kid"] = headers.get("kid")
        headers["jwk"] = jwk_header
        headers["alg"] = alg

        encoded_header = (
            base64.urlsafe_b64encode(bytes(json.dumps(headers), "utf-8"))
            .decode("utf-8")
            .rstrip("=")
        )
        encoded_payload = (
            base64.urlsafe_b64encode(bytes(json.dumps(payload), "utf-8"))
            .decode("utf-8")
            .rstrip("=")
        )

        signature_input = f"{encoded_header}.{encoded_payload}".encode("utf-8")
        hashed_input = Hash.SHA256.new(signature_input)
        signature = PKCS1_15.new(private_key).sign(hashed_input)

        token = f"{encoded_header}.{encoded_payload}.{base64.urlsafe_b64encode(signature).decode('utf-8').replace('=', '')}"

        self.state = token
        return self

    @ChepyDecorators.call_stack
    def rc4_encrypt(self, key: str, key_format: str = "hex") -> EncryptionEncodingT:
        """Encrypt raw state with RC4

        Args:
            key (str): Required. Secret key
            key_format (str, optional): Key format. Defaults to "hex".

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").rc4_encrypt("736563726574").o
            b"9e59bf79a2c0b7d253"
        """
        if isinstance(key, str):
            key = key.encode()
        if key_format == "hex":
            key = binascii.unhexlify(key)
        elif key_format == "base64":
            key = base64.b64decode(key)
        elif key_format == "utf-16-le":
            key = key.decode().encode("utf-16-le")
        elif key_format == "utf-16-be":
            key = key.decode().encode("utf-16-be")
        cipher = ARC4.new(key)
        self.state = binascii.hexlify(cipher.encrypt(self._convert_to_bytes()))
        return self

    @ChepyDecorators.call_stack
    def rc4_decrypt(self, key: str, key_format: str = "hex") -> EncryptionEncodingT:
        """Decrypt raw state with RC4

        Args:
            key (str): Required. Secret key
            key_format (str, optional): Key format. Defaults to "hex".

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("9e59bf79a2c0b7d253").hex_to_str().rc4_decrypt("secret").o
            b"some data"
        """
        if isinstance(key, str):
            key = key.encode()
        if key_format == "hex":
            key = binascii.unhexlify(key)
        elif key_format == "base64":
            key = base64.b64decode(key)
        elif key_format == "utf-16-le":
            key = key.decode().encode("utf-16-le")
        elif key_format == "utf-16-be":
            key = key.decode().encode("utf-16-be")
        cipher = ARC4.new(key)
        self.state = cipher.decrypt(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def des_encrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        key_format: str = "hex",
        iv_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Encrypt raw state with DES

        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            iv_format (str, optional): Format of IV. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").des_encrypt("70617373776f7264").o
            b"1ee5cb52954b211d1acd6e79c598baac"

            To encrypt using a different mode

            >>> Chepy("some data").des_encrypt("password", mode="CTR").o
            b"0b7399049b0267d93d"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, key_format, iv_format)

        if mode == "CBC":
            cipher = DES.new(key, mode=DES.MODE_CBC, iv=iv)
            self.state = cipher.encrypt(Padding.pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "ECB":
            cipher = DES.new(key, mode=DES.MODE_ECB)
            self.state = cipher.encrypt(Padding.pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "CTR":
            cipher = DES.new(key, mode=DES.MODE_CTR, nonce=b"")
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = DES.new(key, mode=DES.MODE_OFB, iv=iv)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def des_decrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        key_format: str = "hex",
        iv_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Decrypt raw state encrypted with DES.

        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            iv_format (str, optional): Format of IV. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("1ee5cb52954b211d1acd6e79c598baac").hex_to_str().des_decrypt("password").o
            b"some data"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, key_format, iv_format)

        if mode == "CBC":
            cipher = DES.new(key, mode=DES.MODE_CBC, iv=iv)
            self.state = Padding.unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "ECB":
            cipher = DES.new(key, mode=DES.MODE_ECB)
            self.state = Padding.unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "CTR":
            cipher = DES.new(key, mode=DES.MODE_CTR, nonce=b"")
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = DES.new(key, mode=DES.MODE_OFB, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def salsa20_encrypt(
        self,
        key: str,
        nonce: str = "0000000000000000",
        key_format: str = "hex",
        nonce_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Encrypt raw state with Salsa 20 rounds

        Args:
            key (str): Required. The secret key
            nonce (str, optional): Nonce. Defaults to '0000000000000000'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            nonce_format (str, optional): Format of nonce. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.
        """

        key, nonce = self._convert_key(key, nonce, key_format, nonce_format)

        cipher = Salsa20.new(key=key, nonce=nonce)
        self.state = cipher.encrypt(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def salsa20_decrypt(
        self,
        key: str,
        nonce: str = "0000000000000000",
        key_format: str = "hex",
        nonce_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Decrypt raw state encrypted with ChaCha 20 rounds.

        Args:
            key (str): Required. The secret key
            nonce (str, optional): nonce for certain modes only. Defaults to '0000000000000000'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            nonce_format (str, optional): Format of nonce. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.
        """

        key, nonce = self._convert_key(key, nonce, key_format, nonce_format)

        cipher = Salsa20.new(key=key, nonce=nonce)
        self.state = cipher.decrypt(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def chacha_encrypt(
        self,
        key: str,
        nonce: str = "0000000000000000",
        key_format: str = "hex",
        nonce_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Encrypt raw state with ChaCha 20 rounds

        Args:
            key (str): Required. The secret key
            nonce (str, optional): Nonce. Defaults to '0000000000000000'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            nonce_format (str, optional): Format of nonce. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.
        """

        key, nonce = self._convert_key(key, nonce, key_format, nonce_format)

        cipher = ChaCha20.new(key=key, nonce=nonce)
        self.state = cipher.encrypt(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def chacha_decrypt(
        self,
        key: str,
        nonce: str = "0000000000000000",
        key_format: str = "hex",
        nonce_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Decrypt raw state encrypted with ChaCha 20 rounds.

        Args:
            key (str): Required. The secret key
            nonce (str, optional): nonce for certain modes only. Defaults to '0000000000000000'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            nonce_format (str, optional): Format of nonce. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.
        """

        key, nonce = self._convert_key(key, nonce, key_format, nonce_format)

        cipher = ChaCha20.new(key=key, nonce=nonce)
        self.state = cipher.decrypt(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def triple_des_encrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        key_format: str = "hex",
        iv_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Encrypt raw state with Triple DES

        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            iv_format (str, optional): Format of IV. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").triple_des_encrypt("super secret password !!", mode="ECB").o
            b"f8b27a0d8c837edc8fb00ea85f502fb4"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, key_format, iv_format)

        if mode == "CBC":
            cipher = DES3.new(key, mode=DES3.MODE_CBC, iv=iv)
            self.state = cipher.encrypt(Padding.pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "ECB":
            cipher = DES3.new(key, mode=DES3.MODE_ECB)
            self.state = cipher.encrypt(Padding.pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "CTR":
            cipher = DES3.new(key, mode=DES3.MODE_CTR, nonce=b"")
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = DES3.new(key, mode=DES3.MODE_OFB, iv=iv)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def triple_des_decrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        key_format: str = "hex",
        iv_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Decrypt raw state encrypted with DES.

        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            iv_format (str, optional): Format of IV. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("f8b27a0d8c837edce87dd13a1ab41f96")
            >>> c.hex_to_str()
            >>> c.triple_des_decrypt("super secret password !!")
            >>> c.o
            b"some data"
        """

        key, iv = self._convert_key(key, iv, key_format, iv_format)

        if mode == "CBC":
            cipher = DES3.new(key, mode=DES3.MODE_CBC, iv=iv)
            self.state = Padding.unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "CBC/NoPadding":
            cipher = DES3.new(key, mode=DES3.MODE_CBC, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "ECB":
            cipher = DES3.new(key, mode=DES3.MODE_ECB)
            self.state = Padding.unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "ECB/NoPadding":
            cipher = DES3.new(key, mode=DES3.MODE_ECB)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "CTR":
            cipher = DES3.new(key, mode=DES3.MODE_CTR, nonce=b"")
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = DES3.new(key, mode=DES3.MODE_OFB, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        else:  # pragma: no cover
            raise ValueError("Invalid mode")

    @ChepyDecorators.call_stack
    def aes_encrypt(
        self,
        key: str,
        iv: str = "00000000000000000000000000000000",
        mode: str = "CBC",
        key_format: str = "hex",
        iv_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Encrypt raw state with AES.
        CFB mode reflects Cyberchef and not native python behaviour.

        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only.
                Defaults to '00000000000000000000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            iv_format (str, optional): Format of IV. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").aes_encrypt("secret password!", mode="ECB").o
            b"5fb8c186394fc399849b89d3b6605fa3"
        """

        assert mode in ["CBC", "CFB", "OFB", "CTR", "ECB", "GCM"], "Not a valid mode."

        key, iv = self._convert_key(key, iv, key_format, iv_format)

        if mode == "CBC":
            cipher = AES.new(key, mode=AES.MODE_CBC, iv=iv)
            self.state = cipher.encrypt(Padding.pad(self._convert_to_bytes(), 16))
            return self
        elif mode == "CFB":
            cipher = AES.new(key, mode=AES.MODE_CFB, iv=iv, segment_size=128)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "ECB":
            cipher = AES.new(key, mode=AES.MODE_ECB)
            self.state = cipher.encrypt(Padding.pad(self._convert_to_bytes(), 16))
            return self
        elif mode == "CTR":
            counter = Counter.new(128, initial_value=int.from_bytes(iv, "big"))
            cipher = AES.new(key, mode=AES.MODE_CTR, counter=counter)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "GCM":
            cipher = AES.new(
                key,
                mode=AES.MODE_GCM,
                nonce=binascii.unhexlify("00000000000000000000000000000000"),
            )
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = AES.new(key, mode=AES.MODE_OFB, iv=iv)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def aes_decrypt(
        self,
        key: str,
        iv: str = "00000000000000000000000000000000",
        mode: str = "CBC",
        key_format: str = "hex",
        iv_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Decrypt raw state encrypted with AES.
        CFB mode reflects Cyberchef and not native python behaviour.

        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only.
                Defaults to '00000000000000000000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            iv_format (str, optional): Format of IV. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("5fb8c186394fc399849b89d3b6605fa3")
            >>> c.hex_to_str()
            >>> c.aes_decrypt("7365637265742070617373776f726421")
            >>> c.o
            b"some data"
        """

        key, iv = self._convert_key(key, iv, key_format, iv_format)

        if mode == "CBC":
            cipher = AES.new(key, mode=AES.MODE_CBC, iv=iv)
            self.state = Padding.unpad(cipher.decrypt(self._convert_to_bytes()), 16)
            return self
        elif mode == "CBC/NoPadding":
            cipher = AES.new(key, mode=AES.MODE_CBC, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "CFB":
            cipher = AES.new(key, mode=AES.MODE_CFB, iv=iv, segment_size=128)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "ECB":
            cipher = AES.new(key, mode=AES.MODE_ECB)
            self.state = Padding.unpad(cipher.decrypt(self._convert_to_bytes()), 16)
            return self
        elif mode == "ECB/NoPadding":
            cipher = AES.new(key, mode=AES.MODE_ECB)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "CTR":
            counter = Counter.new(128, initial_value=int.from_bytes(iv, "big"))
            cipher = AES.new(key, mode=AES.MODE_CTR, counter=counter)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "GCM":
            cipher = AES.new(
                key,
                mode=AES.MODE_GCM,
                nonce=binascii.unhexlify("00000000000000000000000000000000"),
            )
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = AES.new(key, mode=AES.MODE_OFB, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        else:  # pragma: no cover
            raise ValueError("Invalid AES mode")

    @ChepyDecorators.call_stack
    def blowfish_encrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        key_format: str = "hex",
        iv_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Encrypt raw state with Blowfish

        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            iv_format (str, optional): Format of IV. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").blowfish_encrypt("password", mode="ECB").o
            b"d9b0a79853f139603951bff96c3d0dd5"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, key_format, iv_format)

        if mode == "CBC":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_CBC, iv=iv)
            self.state = cipher.encrypt(Padding.pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "ECB":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_ECB)
            self.state = cipher.encrypt(Padding.pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "CTR":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_CTR, nonce=b"")
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_OFB, iv=iv)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def blowfish_decrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        key_format: str = "hex",
        iv_format: str = "hex",
    ) -> EncryptionEncodingT:
        """Encrypt raw state with Blowfish

        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only.
                Defaults to '00000000000000000000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            key_format (str, optional): Format of key. Defaults to 'hex'.
            iv_format (str, optional): Format of IV. Defaults to 'hex'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("d9b0a79853f13960fcee3cae16e27884")
            >>> c.hex_to_str()
            >>> c.blowfish_decrypt("password", key_format="utf-8")
            >>> c.o
            b"some data"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, key_format, iv_format)

        if mode == "CBC":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_CBC, iv=iv)
            self.state = Padding.unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "ECB":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_ECB)
            self.state = Padding.unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "CTR":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_CTR, nonce=b"")
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_OFB, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def vigenere_encode(self, key: str) -> EncryptionEncodingT:
        """Vigenere encode

        Args:
            key (str): Key

        Raises:
            ValueError: Key is not alpha
            ValueError: Key is not provided

        Returns:
            Chepy: The Chepy object.
        """
        input_str = self._convert_to_str()
        alphabet = "abcdefghijklmnopqrstuvwxyz"
        key = key.lower()
        output = ""
        fail = 0

        if not key:
            raise ValueError("No key entered")  # pragma: no cover
        if not key.isalpha():
            raise ValueError("The key must consist only of letters")  # pragma: no cover

        for i in range(len(input_str)):
            if input_str[i].isalpha() and input_str[i].lower() in alphabet:
                is_upper = input_str[i].isupper()
                input_char = input_str[i].lower()
                key_char = key[(i - fail) % len(key)]
                key_index = alphabet.index(key_char)
                input_index = alphabet.index(input_char)
                encoded_index = (key_index + input_index) % 26
                encoded_char = alphabet[encoded_index]
                output += encoded_char.upper() if is_upper else encoded_char
            else:
                output += input_str[i]
                fail += 1

        self.state = output
        return self

    @ChepyDecorators.call_stack
    def vigenere_decode(self, key: str) -> EncryptionEncodingT:
        """Vigenere decode

        Args:
            key (str): Key

        Raises:
            ValueError: Key is not alpha
            ValueError: Key is not provided

        Returns:
            Chepy: The Chepy object.
        """
        input_str = self._convert_to_str()
        alphabet = "abcdefghijklmnopqrstuvwxyz"
        output = ""
        fail = 0
        key = key.lower()

        if not key:
            raise ValueError("No key entered")  # pragma: no cover
        if not key.isalpha():
            raise ValueError("The key must consist only of letters")  # pragma: no cover

        for i in range(len(input_str)):
            if input_str[i].isalpha() and input_str[i].lower() in alphabet:
                is_upper = input_str[i].isupper()
                input_char = input_str[i].lower()
                key_char = key[(i - fail) % len(key)]
                key_index = alphabet.index(key_char)
                input_index = alphabet.index(input_char)
                encoded_index = (input_index - key_index + len(alphabet)) % len(
                    alphabet
                )
                encoded_char = alphabet[encoded_index]
                output += encoded_char.upper() if is_upper else encoded_char
            else:
                output += input_str[i]
                fail += 1

        self.state = output
        return self

    @ChepyDecorators.call_stack
    def affine_encode(self, a: int = 1, b: int = 1) -> EncryptionEncodingT:
        """Encode with Affine cipher

        Args:
            a (int, optional): Multiplier value. Defaults to 1
            b (int, optional): Additive value. Defaults to 1

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("secret").affine_encode().o
            "TFDSFU"
        """
        self.state = pycipher.Affine(a=a, b=b).encipher(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def affine_decode(self, a: int = 1, b: int = 1) -> EncryptionEncodingT:
        """Decode Affine cipher

        Args:
            a (int, optional): Multiplier value. Defaults to 1
            b (int, optional): Additive value. Defaults to 1

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("TFDSFU").affine_decode().o
            "SECRET"
        """
        self.state = pycipher.Affine(a=a, b=b).decipher(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def atbash(self) -> EncryptionEncodingT:
        """Encode with Atbash cipher

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("secret").atbash_encode().o
            "HVXIVG"
        """
        # Reference https://github.com/jameslyons/pycipher/blob/master/pycipher/atbash.py
        key = "ZYXWVUTSRQPONMLKJIHGFEDCBA"
        data = self._convert_to_str()
        ret = ""
        arr = Ciphers.ATBASH
        for c in data:
            if c.isalpha():
                if c.islower():
                    ret += key[arr[c.upper()]].lower()
                else:
                    ret += key[arr[c]]
            else:
                ret += c
        self.state = ret
        # self.state = pycipher.Atbash().encipher(self._convert_to_str(), keep_punct=True)
        return self

    @ChepyDecorators.call_stack
    def to_morse_code(
        self,
        dot: str = ".",
        dash: str = "-",
        letter_delim: str = " ",
        word_delim: str = "\n",
    ) -> EncryptionEncodingT:
        """Encode string to morse code

        Args:
            dot (str, optional): The char for dot. Defaults to ".".
            dash (str, optional): The char for dash. Defaults to "-".
            letter_delim (str, optional): Letter delimiter. Defaults to " ".
            word_delim (str, optional): Word delimiter. Defaults to "\\n".

        Returns:
            Chepy: The Chepy object.
        """
        encode = ""
        morse_code_dict = EncryptionConsts.MORSE_CODE_DICT
        for k, v in morse_code_dict.items():
            morse_code_dict[k] = v.replace(".", dot).replace("-", dash)
        for word in self._convert_to_str().split():
            for w in word:
                encode += morse_code_dict.get(w.upper(), "") + letter_delim
            encode += word_delim
        self.state = encode
        return self

    @ChepyDecorators.call_stack
    def from_morse_code(
        self,
        dot: str = ".",
        dash: str = "-",
        letter_delim: str = " ",
        word_delim: str = "\n",
    ) -> EncryptionEncodingT:
        """Decode morse code

        Args:
            dot (str, optional): The char for dot. Defaults to ".".
            dash (str, optional): The char for dash. Defaults to "-".
            letter_delim (str, optional): Letter delimiter. Defaults to " ".
            word_delim (str, optional): Word delimiter. Defaults to "\\n".

        Returns:
            Chepy: The Chepy object.
        """
        decode = ""
        morse_code_dict = EncryptionConsts.MORSE_CODE_DICT
        for k, v in morse_code_dict.items():
            morse_code_dict[k] = v.replace(".", dot).replace("-", dash)

        morse_code_dict = {value: key for key, value in morse_code_dict.items()}
        for chars in self._convert_to_str().split(letter_delim):
            if word_delim in chars:
                chars = re.sub(word_delim, "", chars, re.I)
                if morse_code_dict.get(chars) is not None:
                    decode += " " + morse_code_dict.get(chars, "")
                else:  # pragma: no cover
                    decode += " " + chars
            else:
                decode += morse_code_dict.get(chars, "")
        self.state = decode
        return self

    @ChepyDecorators.call_stack
    def rsa_encrypt(
        self,
        public_key: str,
        is_file: bool = True,
        passphrase: str = None,
        cipher: Literal["OAEP", "PKCS"] = "OAEP",
    ) -> EncryptionEncodingT:
        """Encrypt data with RSA Public key in PEM format

        Args:
            public_key (str): Path to Public key
            is_file (bool): If supplied argument is a PEM file path. Defaults to false
            passphrase (str): passphrase. Defaults to None
            cipher (str): Cipher type. Defaults to OAEP

        Returns:
            Chepy: The Chepy object
        """
        rsa = self._rsa_process_key(public_key, is_file, passphrase)
        c = self._rsa_cipher(cipher, rsa)
        self.state = c.encrypt(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def rsa_decrypt(
        self,
        private_key: str,
        is_file: bool = True,
        passphrase: str = None,
        cipher: Literal["OAEP", "PKCS"] = "OAEP",
    ) -> EncryptionEncodingT:
        """Decrypt data with RSA Private key in PEM format

        Args:
            private_key (str): Path to Private key
            is_file (bool): If supplied argument is a PEM file path. Defaults to false
            passphrase (str): passphrase. Defaults to None
            cipher (str): Cipher type. Defaults to OAEP

        Returns:
            Chepy: The Chepy object
        """
        rsa = self._rsa_process_key(private_key, is_file, passphrase)
        c = self._rsa_cipher(cipher, rsa)
        if cipher == "PKCS":
            self.state = c.decrypt(self._convert_to_bytes(), None)
        else:
            self.state = c.decrypt(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def rsa_sign(
        self,
        private_key: str,
        is_file: bool = True,
        passphrase: str = None,
        hash_format: Literal["SHA256", "SHA512", "SHA1", "MD5", "SHA384"] = "SHA256",
    ) -> EncryptionEncodingT:
        """Sign data in state with RSA Private key in PEM format

        Args:
            private_key (str): Private key
            is_file (bool): If supplied argument is a PEM file path. Defaults to false
            passphrase (str): passphrase. Defaults to None
            hash_format (str): hash type. Defaults to SHA256

        Returns:
            Chepy: The Chepy object
        """
        rsa = self._rsa_process_key(private_key, is_file, passphrase)
        h = getattr(Hash, hash_format).new(self._convert_to_bytes())
        self.state = PKCS1_15.new(rsa).sign(h)
        return self

    @ChepyDecorators.call_stack
    def rsa_verify(
        self,
        signature: bytes,
        public_key: str,
        is_file: bool = True,
        passphrase: str = None,
        hash_format: Literal["SHA256", "SHA512", "SHA1", "MD5", "SHA384"] = "SHA256",
    ) -> EncryptionEncodingT:  # pragma: no cover
        """Verify data in state with RSA Public key in PEM format

        Args:
            signature (bytes): The signature as bytes
            public_key (str): Path to Private key
            is_file (bool): If supplied argument is a PEM file path. Defaults to false
            passphrase (str): passphrase. Defaults to None
            hash_format (str): Cipher type. Defaults to SHA256

        Returns:
            Chepy: The Chepy object
        """
        rsa = self._rsa_process_key(public_key, is_file, passphrase)
        h = getattr(Hash, hash_format).new(self._convert_to_bytes())
        self.state = PKCS1_15.new(rsa).verify(h, signature)
        return self

    @ChepyDecorators.call_stack
    def rsa_private_pem_to_jwk(self) -> EncryptionEncodingT:
        """Convert RSA PEM private key to jwk format

        Returns:
            Chepy: The Chepy object.
        """
        # Load the PEM private key
        private_key = RSA.import_key(self._convert_to_str())

        n = private_key.n
        e = private_key.e
        d = private_key.d
        p = private_key.p
        q = private_key.q
        dp = private_key.d % (p - 1)
        dq = private_key.d % (q - 1)
        qi = pow(q, -1, p)

        n_base64url = (
            base64.urlsafe_b64encode(n.to_bytes((n.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("=")
        )
        e_base64url = (
            base64.urlsafe_b64encode(e.to_bytes((e.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("=")
        )
        d_base64url = (
            base64.urlsafe_b64encode(d.to_bytes((d.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("=")
        )
        p_base64url = (
            base64.urlsafe_b64encode(p.to_bytes((p.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("=")
        )
        q_base64url = (
            base64.urlsafe_b64encode(q.to_bytes((q.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("=")
        )
        dp_base64url = (
            base64.urlsafe_b64encode(dp.to_bytes((dp.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("=")
        )
        dq_base64url = (
            base64.urlsafe_b64encode(dq.to_bytes((dq.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("=")
        )
        qi_base64url = (
            base64.urlsafe_b64encode(qi.to_bytes((qi.bit_length() + 7) // 8, "big"))
            .decode("utf-8")
            .rstrip("=")
        )

        private = {
            "p": p_base64url,
            "kty": "RSA",
            "q": q_base64url,
            "d": d_base64url,
            "e": e_base64url,
            "qi": qi_base64url,
            "dp": dp_base64url,
            "dq": dq_base64url,
            "n": n_base64url,
        }

        public = {"kty": "RSA", "e": e_base64url, "n": n_base64url}
        self.state = {"private": private, "public": public}
        return self

    @ChepyDecorators.call_stack
    def rsa_public_key_from_jwk(self) -> EncryptionEncodingT:
        """Generate RSA public key in PEM format from JWK

        Raises:
            AssertionError: If n or e not found

        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, dict), "State should be a dict"
        jwk_key = self.state
        if not "e" in jwk_key or "n" not in jwk_key:
            raise AssertionError("e or n not found")  # pragma: no cover
        e = int.from_bytes(base64.urlsafe_b64decode(jwk_key["e"] + "=="), "big")
        n = int.from_bytes(base64.urlsafe_b64decode(jwk_key["n"] + "=="), "big")

        public_key = RSA.construct((n, e))

        self.state = public_key.export_key().decode("utf-8")
        return self

    @ChepyDecorators.call_stack
    def monoalphabetic_substitution(
        self, mapping: Dict[str, str] = {}
    ) -> EncryptionEncodingT:
        """Monoalphabetic substitution. Re-map characters

        Args:
            mapping (Dict[str, str], optional): Mapping of characters where key is the character to map and value is the new character to replace with. Defaults to {}.

        Returns:
            Chepy: The Chepy object
        """
        hold = ""
        cipher = self._convert_to_str()
        for c in cipher:
            hold += mapping.get(c.lower(), c)
        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def to_letter_number_code(
        self, join_by: Union[str, bytes] = b" "
    ) -> EncryptionEncodingT:
        """Encode to A1Z26

        Args:
            join_by (Union[str, bytes], optional): join output by. Defaults to b' '.

        Returns:
            Chepy: The Chepy object.
        """
        join_by = self._str_to_bytes(join_by)
        data = list(self._convert_to_str())
        hold = []
        for i, d in enumerate(data):
            hold.append(f"{d}{i}".encode())
        random.shuffle(hold)
        self.state = join_by.join(hold)
        return self

    @ChepyDecorators.call_stack
    def from_letter_number_code(
        self, delimiter: Union[str, bytes] = " ", join_by: Union[str, bytes] = ""
    ) -> EncryptionEncodingT:
        """Decode A1Z26

        Args:
            delimiter (Union[str, bytes], optional): Split on. Defaults to ' '.
            join_by (Union[str, bytes], optional): Join output by. Defaults to ''.

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_str()
        delimiter = detect_delimiter(data, default_delimiter=delimiter)
        data = data.split(delimiter)
        hold = ["§" for _ in range(len(data))]
        for d in data:
            try:
                hold[int(d[1:])] = d[0]
            except:  # pragma: no cover
                continue
        self.state = join_by.join(hold).encode()
        return self

    @ChepyDecorators.call_stack
    def ls47_encrypt(
        self, password: str, padding: int = 10, signature: str = ""
    ) -> EncryptionEncodingT:
        """LS47 encrypt

        Args:
            password (str): password
            padding (int, optional): Padding. Defaults to 10.
            signature (str, optional): Signature to prepend. Defaults to ''.

        Returns:
            Chepy: The Chepy object.
        """
        key = _derive_key(password)
        self.state = _ls47_enc(
            key, self._convert_to_str(), padding_size=padding, signature=signature
        )
        return self

    @ChepyDecorators.call_stack
    def ls47_decrypt(self, password: str, padding: int = 10) -> EncryptionEncodingT:
        """LS47 decrypt

        Args:
            password (str): password
            padding (int, optional): Padding. Defaults to 10.

        Returns:
            Chepy: The Chepy object.
        """
        key = _derive_key(password)
        self.state = _ls47_dec(key, padding, self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def bifid_encode(self, key: Union[bytes, str] = "") -> EncryptionEncodingT:
        """Bifid / polybius decode

        Args:
            key (Union[str, bytes], optional): Key. Defaults to "".

        Returns:
            Chepy: The Chepy object.
        """
        key = self._bytes_to_str(key)
        key = "".join(re.findall(r"[A-Z]+", key))
        keyword_str = key.upper().replace("J", "I")
        # keyword_set = set(keyword_str)
        # keyword_list = []
        alpha = "ABCDEFGHIKLMNOPQRSTUVWXYZ"
        x_co = []
        y_co = []
        structure = []

        output = ""
        count = 0

        polybius = Ciphers.gen_polybius_square(keyword_str)

        for letter in self._convert_to_str().replace("J", "I"):
            alp_ind = letter.upper() in alpha
            pol_ind = -1

            if alp_ind:
                for i in range(5):
                    if letter.upper() in polybius[i]:
                        pol_ind = polybius[i].index(letter.upper())
                        x_co.append(pol_ind)
                        y_co.append(i)

                if letter in alpha:
                    structure.append(True)  # pragma: no cover
                elif alp_ind:
                    structure.append(False)
            else:
                structure.append(letter)  # pragma: no cover

        trans = "".join(map(str, y_co)) + "".join(map(str, x_co))

        for pos in structure:
            if isinstance(pos, bool):
                coords = trans[2 * count : 2 * count + 2]
                coords = list(map(int, coords))

                output += (
                    polybius[coords[0]][coords[1]]
                    if pos
                    else polybius[coords[0]][coords[1]].lower()
                )
                count += 1
            else:
                output += pos  # pragma: no cover

        self.state = output
        return self

    @ChepyDecorators.call_stack
    def bifid_decode(self, key: Union[str, bytes] = ""):
        """Bifid / polybius decode

        Args:
            key (Union[str, bytes], optional): Key. Defaults to "".

        Returns:
            Chepy: The Chepy object.
        """
        key = self._bytes_to_str(key)
        key = "".join(re.findall(r"[A-Z]+", key))
        keyword_str = key.upper().replace("J", "I")
        # keyword_set = set(keyword_str)
        alpha = "ABCDEFGHIKLMNOPQRSTUVWXYZ"
        structure = []

        output = ""
        count = 0
        trans = ""

        polybius = Ciphers.gen_polybius_square(keyword_str)

        for letter in self._convert_to_str().replace("J", "I"):
            alp_ind = letter.upper() in alpha
            pol_ind = -1

            if alp_ind:
                for i in range(5):
                    if letter.upper() in polybius[i]:
                        pol_ind = polybius[i].index(letter.upper())
                        trans += f"{i}{pol_ind}"

                if letter in alpha:
                    structure.append(True)  # pragma: no cover
                elif alp_ind:
                    structure.append(False)
            else:
                structure.append(letter)  # pragma: no cover

        for pos in structure:
            if isinstance(pos, bool):
                coords = [int(trans[count]), int(trans[count + len(trans) // 2])]

                output += (
                    polybius[coords[0]][coords[1]]
                    if pos
                    else polybius[coords[0]][coords[1]].lower()
                )
                count += 1
            else:
                output += pos  # pragma: no cover

        self.state = output
        return self

    @ChepyDecorators.call_stack
    def huffman_encode(self) -> EncryptionEncodingT:
        """Huffman encode

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_str()
        root = Ciphers.build_huffman_tree(data)
        huffman_codes = {}
        Ciphers.build_huffman_codes(root, "", huffman_codes)
        encoded_data = "".join(huffman_codes[char] for char in data)
        self.state = {"encoded": encoded_data, "codes": huffman_codes}
        return self

    @ChepyDecorators.call_stack
    def huffman_decode(self, huffman_codes: Dict[str, str]) -> EncryptionEncodingT:
        """Huffman decode

        Args:
            huffman_codes (Dict[str, str]): Huffman codes as a dict

        Returns:
            Chepy: The Chepy object.
        """
        decoded_data = ""
        current_code = ""

        encoded_data = self._convert_to_str()
        for bit in encoded_data:
            current_code += bit
            for char, code in huffman_codes.items():
                if code == current_code:
                    decoded_data += char
                    current_code = ""
                    break

        self.state = decoded_data
        return self

    @ChepyDecorators.call_stack
    def cetacean_encode(self) -> EncryptionEncodingT:
        """Cetacean encode

        Returns:
            Chepy: The Chepy object.
        """
        result = []
        charArray = list(self._convert_to_str())

        for character in charArray:
            if character == " ":
                result.append(character)
            else:
                binaryArray = format(ord(character), "016b")
                result.append(
                    "".join(["e" if bit == "1" else "E" for bit in binaryArray])
                )

        self.state = "".join(result)
        return self

    @ChepyDecorators.call_stack
    def cetacean_decode(self) -> EncryptionEncodingT:
        """Cetacean decode

        Returns:
            Chepy: The Chepy object.
        """
        binaryArray = []

        for char in self._convert_to_str():
            if char == " ":
                binaryArray.extend([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0])
            else:
                binaryArray.append(1 if char == "e" else 0)

        byteArray = []

        for i in range(0, len(binaryArray), 16):
            byteArray.append("".join(map(str, binaryArray[i : i + 16])))

        self.state = "".join([chr(int(byte, 2)) for byte in byteArray])
        return self

    @ChepyDecorators.call_stack
    def rabbit(self, key: str, iv: Union[None, str] = None) -> EncryptionEncodingT:
        """Rabbit encryption/decryption

        Args:
            key (str): Key
            iv (Union[None,str], optional): IV. Defaults to None.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = Rabbit(key, iv).encrypt(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def fernet_encrypt(
        self, key: Union[bytes, str], encode_key: bool = False
    ) -> EncryptionEncodingT:
        """Fernet encrypt

        Args:
            key (Union[bytes, str]): Key to encrypt with. This should be 32 bytes long
            encode_key (bool, optional): If key should be base64 encoded. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        key = self._str_to_bytes(key)
        if encode_key:
            key = base64.b64encode(key)
        out = Fernet(key).encrypt(self._convert_to_bytes())
        self.state = out
        return self

    @ChepyDecorators.call_stack
    def fernet_decrypt(
        self, key: Union[bytes, str], encode_key: bool = False
    ) -> EncryptionEncodingT:
        """Fernet decrypt

        Args:
            key (Union[bytes, str]): Key to encrypt with. This should be 32 bytes long
            encode_key (bool, optional): If key should be base64 encoded. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        key = self._str_to_bytes(key)
        if encode_key:
            key = base64.b64encode(key)
        out = Fernet(key).decrypt(self._convert_to_bytes())
        self.state = out
        return self

    @ChepyDecorators.call_stack
    def railfence_encode(self, key=2, offset=0) -> EncryptionEncodingT:
        """Encode to railfence

        Args:
            key (int, optional): Key. Should be equal or larger than data. Defaults to 2.
            offset (int, optional): Offset. Defaults to 0.

        Returns:
            Chepy: The Chepy object.
        """
        key, offset = int(key), int(offset)
        data = self._convert_to_str()
        if key < 2:
            raise ValueError("Key has to be bigger than 2")  # pragma: no cover
        elif key > len(data):
            raise ValueError(
                "Key should be smaller than the plain text's length"
            )  # pragma: no cover

        if offset < 0:
            raise ValueError("Offset has to be a positive integer")  # pragma: no cover

        cycle = (key - 1) * 2
        rows = [""] * key

        for pos in range(len(data)):
            row_idx = key - 1 - abs(cycle // 2 - (pos + offset) % cycle)
            rows[row_idx] += data[pos]

        self.state = "".join(rows).strip()
        return self

    @ChepyDecorators.call_stack
    def railfence_decode(self, key=2, offset=0) -> EncryptionEncodingT:
        """Decode railfence

        Args:
            key (int, optional): Key. Should be equal or larger than data. Defaults to 2.
            offset (int, optional): Offset. Defaults to 0.

        Returns:
            Chepy: The Chepy object.
        """
        key, offset = int(key), int(offset)
        cipher = self._convert_to_str()

        if key < 2:
            raise ValueError("Key has to be bigger than 2")  # pragma: no cover
        elif key > len(cipher):
            raise ValueError(
                "Key should be smaller than the cipher's length"
            )  # pragma: no cover

        if offset < 0:
            raise ValueError("Offset has to be a positive integer")  # pragma: no cover

        cycle = (key - 1) * 2
        plaintext = [""] * len(cipher)

        j = 0
        for y in range(key):
            for x in range(len(cipher)):
                if (y + x + offset) % cycle == 0 or (y - x - offset) % cycle == 0:
                    plaintext[x] = cipher[j]
                    j += 1

        self.state = "".join(plaintext).strip()
        return self

    @ChepyDecorators.call_stack
    def gpp_decrypt(self):
        """Decrypt Group Policy Preferences (GPP) password

        Returns:
            Chepy: The Chepy object.
        """
        password = self._convert_to_bytes()
        password = base64.b64decode(password + b"=" * 3)
        key = (
            b"\x4e\x99\x06\xe8\xfc\xb6\x6c\xc9\xfa\xf4\x93\x10\x62\x0f\xfe\xe8"
            b"\xf4\x96\xe8\x06\xcc\x05\x79\x90\x20\x9b\x09\xa4\x33\xb6\x6c\x1b"
        )
        iv = b"\x00" * 16
        aes = AES.new(key, AES.MODE_CBC, iv)
        self.state = Padding.unpad(aes.decrypt(password), 16)
        self.remove_nullbytes()
        return self

    @ChepyDecorators.call_stack
    def pgp_decrypt(
        self, passphrase: Union[str, bytes], armoured: bool = False
    ) -> EncryptionEncodingT:
        """Decrypt PGP encrypted file with passphrase

        Args:
            passphrase (Union[str, bytes]): passphrase
            armoured (bool): PGP armoured format

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_bytes()
        out = pgpy.PGPMessage.decrypt(
            pgpy.PGPMessage.from_blob(data), passphrase=passphrase
        )
        if armoured:
            self.state = out.__str__()
        else:
            self.state = out.message
        return self

    @ChepyDecorators.call_stack
    def pgp_encrypt(
        self, passphrase: Union[str, bytes], armoured=False
    ) -> EncryptionEncodingT:
        """PGP encrypt

        Args:
            passphrase (Union[str, bytes]): passphrase
            armoured (bool, optional): PGP armoured format. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        passphrase = self._to_bytes(passphrase)
        msg = pgpy.PGPMessage.new(self._convert_to_bytes())
        enc = pgpy.PGPMessage.encrypt(msg, passphrase)
        if armoured:
            self.state = pgpy.PGPMessage.from_blob(enc.__bytes__()).__str__()
        else:
            self.state = enc.__bytes__()
        return self

    @ChepyDecorators.call_stack
    def to_zeckendorf(self, space_delim=".", byte_delim="|") -> EncryptionEncodingT:
        """Convert the current state to a Zeckendorf representation.

        Args:
            space_delim (str, optional): Delimiter for space-separated Fibonacci numbers. Defaults to ".".
            byte_delim (str, optional): Delimiter for byte-separated groups. Defaults to "|".

        Returns:
            Chepy: The Chepy object.
        """
        if isinstance(space_delim, str):
            space_delim = space_delim.encode()
        if isinstance(byte_delim, str):
            byte_delim = byte_delim.encode()
        data = self._convert_to_bytes()

        encoded_parts = []

        for byte_val in data:
            if byte_val == 0:
                # Handle zero case - we'll represent it as empty (no Fibonacci numbers)
                encoded_parts.append("")  # pragma: no cover
            else:
                zeck_repr = Zeckendorf.to_zeckendorf(byte_val)
                zeck_str = space_delim.join(str(x).encode() for x in zeck_repr)
                encoded_parts.append(zeck_str)

        self.state = byte_delim.join(encoded_parts)
        return self

    @ChepyDecorators.call_stack
    def from_zeckendorf(self, space_delim=".", byte_delim="|") -> EncryptionEncodingT:
        """Decode a Zeckendorf representation into bytes.

        Args:
            space_delim (str, optional): Delimiter for space-separated Fibonacci numbers. Defaults to ".".
            byte_delim (str, optional): Delimiter for byte-separated groups. Defaults to "|".

        Returns:
            Chepy: The Chepy object.
        """
        if isinstance(space_delim, str):
            space_delim = space_delim.encode()
        if isinstance(byte_delim, str):
            byte_delim = byte_delim.encode()

        data = self._convert_to_bytes()
        parts = data.split(byte_delim)
        decoded_bytes = []

        for part in parts:
            if not part:  # Empty part represents 0
                decoded_bytes.append(0)  # pragma: no cover
            else:
                # Sum the Fibonacci numbers
                fib_nums = [int(s) for s in part.split(space_delim) if s]
                byte_val = sum(fib_nums)
                decoded_bytes.append(byte_val)

        self.state = bytes(decoded_bytes)
        return self

    @ChepyDecorators.call_stack
    def sms_encode_multitap(self):
        """
        Encode text to multi-tap SMS format.
        Example: 'sun' -> '7778866'
        Preserves special characters that aren't on phone keys.

        Returns:
            Chepy: The Chepy object.
        """
        KEY_MAP = {
            "2": "ABC",
            "3": "DEF",
            "4": "GHI",
            "5": "JKL",
            "6": "MNO",
            "7": "PQRS",
            "8": "TUV",
            "9": "WXYZ",
            "0": " ",
        }
        REVERSE_MAP = {}

        for key, letters in KEY_MAP.items():
            for i, letter in enumerate(letters):
                REVERSE_MAP[letter] = key * (i + 1)

        result = []

        for char in self._convert_to_str():
            upper_char = char.upper()

            # Check if character can be encoded
            if upper_char in REVERSE_MAP:
                result.append(REVERSE_MAP[upper_char])
            else:
                # Preserve special characters as-is
                result.append(char)

        self.state = "".join(result)
        return self

    @ChepyDecorators.call_stack
    def sms_decode_multitap(self):
        """
        Decode multi-tap SMS format to text.
        Example: '7778866' -> 'SUN'
        Preserves special characters that aren't numeric keypresses.

        Returns:
            Chepy: The Chepy object.
        """
        KEY_MAP = {
            "2": "ABC",
            "3": "DEF",
            "4": "GHI",
            "5": "JKL",
            "6": "MNO",
            "7": "PQRS",
            "8": "TUV",
            "9": "WXYZ",
            "0": " ",
        }

        result = []
        i = 0

        encoded = self._convert_to_str()

        while i < len(encoded):
            current_char = encoded[i]

            # If it's not a digit, preserve it as-is
            if not current_char.isdigit():
                result.append(current_char)
                i += 1
                continue

            # If it's not a valid key, preserve it
            if current_char not in KEY_MAP:  # pragma: no cover
                result.append(current_char)
                i += 1
                continue

            letters = KEY_MAP[current_char]
            letters_count = len(letters)

            # Count consecutive same digits
            count = 0
            while i < len(encoded) and encoded[i] == current_char:
                count += 1
                i += 1

            # Decode groups of presses, handling wrapping
            while count > 0:
                presses_for_this_letter = count % letters_count
                if presses_for_this_letter == 0:
                    presses_for_this_letter = letters_count

                letter_index = presses_for_this_letter - 1
                result.append(letters[letter_index])
                count -= presses_for_this_letter

        self.state = "".join(result)
        return self



from typing import Literal, TypeVar, Union

from ..core import ChepyCore, ChepyDecorators
import lazy_import

RSA = lazy_import.lazy_module("Crypto.PublicKey.RSA")
ECC = lazy_import.lazy_module("Crypto.PublicKey.ECC")
OpenSSL = lazy_import.lazy_module("OpenSSL")


PublickeyT = TypeVar("PublickeyT", bound="Publickey")


class Publickey(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    def _convert_cert_to_obj(self, cert):
        issuer = cert.get_issuer()
        subject = cert.get_subject()
        pubkey = cert.get_pubkey()
        info = {
            "version": cert.get_version(),
            "serial": cert.get_serial_number(),
            "algo": cert.get_signature_algorithm(),
            "before": cert.get_notBefore(),
            "after": cert.get_notAfter(),
            "issuer": {
                "C": issuer.C,
                "ST": issuer.ST,
                "L": issuer.L,
                "O": issuer.O,
                "OU": issuer.OU,
                "CN": issuer.CN,
                "email": issuer.emailAddress,
            },
            "subject": {
                "C": subject.C,
                "ST": subject.ST,
                "L": subject.L,
                "O": subject.O,
                "OU": subject.OU,
                "CN": subject.CN,
                "email": subject.emailAddress,
            },
            "pubkey": {"bits": pubkey.bits()},
        }
        return info

    @ChepyDecorators.call_stack
    def parse_x509_pem(self) -> PublickeyT:
        """Parse X509 cert in PEM format

        X.509 is an ITU-T standard for a public key infrastructure (PKI)
        and Privilege Management Infrastructure (PMI). It is commonly involved
        with SSL/TLS security. This operation displays the contents of
        a certificate in a human readable format, similar to the openssl command line tool.

        Returns:
            Chepy: A Chepy object.

        Examples:
            >>> Chepy(path).load_file().parse_x509_pem().o
            {
                "version": 0,
                ...
                "after": b"20201101152508Z",
                "issuer": {
                    "C": "US",
                    ...
                    "email": "none@email.com",
                },
                "subject": {
                    "C": "US",
                    ...
                    "email": "none@email.com",
                },
                "pubkey": {"bits": 1024},
            }
        """
        cert = OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_PEM, self._convert_to_str()
        )
        info = self._convert_cert_to_obj(cert)
        self.state = info
        return self

    @ChepyDecorators.call_stack
    def parse_x509_der_hex(self) -> PublickeyT:
        """Parse X509 cert in DER format

        X.509 is an ITU-T standard for a public key infrastructure (PKI)
        and Privilege Management Infrastructure (PMI). It is commonly involved
        with SSL/TLS security.<br><br>This operation displays the contents of
        a certificate in a human readable format, similar to the openssl command line tool.

        Returns:
            Chepy: A Chepy object.
        """
        cert = OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_ASN1, self._convert_to_bytes()
        )
        info = self._convert_cert_to_obj(cert)
        self.state = info
        return self

    @ChepyDecorators.call_stack
    def public_from_x509(self) -> PublickeyT:
        """Get public key from x509 certificate

        Returns:
            Chepy: The Chepy object.
        """
        crt_obj = OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_PEM, self.state
        )
        pub_key_object = crt_obj.get_pubkey()
        pub_key_string = OpenSSL.crypto.dump_publickey(
            OpenSSL.crypto.FILETYPE_PEM, pub_key_object
        )
        self.state = pub_key_string
        return self

    @ChepyDecorators.call_stack
    def pem_to_der_hex(self) -> PublickeyT:
        """Convert PEM cert to DER format

        Converts PEM (Privacy Enhanced Mail) format to a hexadecimal
        DER (Distinguished Encoding Rules) string.

        Returns:
            Chepy: The Chepy object.
        """
        cert_pem = OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_PEM, self.state
        )
        self.state = OpenSSL.crypto.dump_certificate(
            OpenSSL.crypto.FILETYPE_ASN1, cert_pem
        )
        return self

    @ChepyDecorators.call_stack
    def der_hex_to_pem(self) -> PublickeyT:
        """Convert DER format to PEM cert.

        Converts a hexadecimal DER (Distinguished Encoding Rules)
        string into PEM (Privacy Enhanced Mail) format.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy(str(Path().absolute() / "tests/files/test.der"))
            >>> c.load_file()
            >>> c.der_hex_to_pem()
            >>> c.o.decode()
            -----BEGIN CERTIFICATE-----
            MIICeTCCAeICCQDi5dgCpKMeHTANBgkqhkiG9w0BAQsFADCBgDELMAkGA1UEBhMC
            VVMxDDAKBgNVBAgMA2xvbDEMMAoGA1UEBwwDbnljMRIwEAYDVQQKDAlzZWN1cmlz
            ...
            wQSFm54UNQ/vjM12yZ+C5c3268Vo8jSP7mI5R3wn6XztjUSXkDg5/3IL3kojti/h
            nyhBHx2QCVke7BxWw3HWkbZ/1BKl0HnCGyd5HDTuOtlBmTS+QrJoNpdsn0zq4fvc
            igbV1IJdKTBAiZzaOQ==
            -----END CERTIFICATE-----
        """
        cert_pem = OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_ASN1, self._convert_to_bytes()
        )
        self.state = OpenSSL.crypto.dump_certificate(
            OpenSSL.crypto.FILETYPE_PEM, cert_pem
        )
        return self

    @ChepyDecorators.call_stack
    def parse_public_pem(self) -> PublickeyT:
        """Parse pubkey PEM to get n and e

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("tests/files/public.pem").load_file().parse_public_pem().get_by_key("e").o
            65537
        """
        key = RSA.importKey(self.state)
        self.state = {"n": key.n, "e": key.e}
        return self

    @ChepyDecorators.call_stack
    def parse_private_pem(self) -> PublickeyT:
        """Parse private key PEM

        Returns:
            Chepy: The Chepy object.
        """
        key = RSA.importKey(self.state)
        self.state = {
            "d": key.d,
            "e": key.e,
            "n": key.n,
            "p": key.p,
            "q": key.q,
            "u": key.u,
        }
        return self

    @ChepyDecorators.call_stack
    def dump_pkcs12_cert(self, password: Union[str, bytes]) -> PublickeyT:
        """Get the private key and cert from pkcs12 cert

        Args:
            password (str): Password for certificate

        Returns:
            Chepy: The Chepy object.
        """
        password = self._to_bytes(password)
        pk12 = OpenSSL.crypto.load_pkcs12(self._convert_to_bytes(), password)
        self.state = {
            "private": OpenSSL.crypto.dump_privatekey(
                OpenSSL.crypto.FILETYPE_PEM, pk12.get_privatekey()
            ),
            "cert": OpenSSL.crypto.dump_certificate(
                OpenSSL.crypto.FILETYPE_PEM, pk12.get_certificate()
            ),
        }
        return self

    @ChepyDecorators.call_stack
    def generate_rsa_keypair(
        self,
        bits: int = 1024,
        format: Literal["PEM", "DER"] = "PEM",
        passphrase: str = None,
    ) -> PublickeyT:
        """Generate RSA key pair

        Args:
            bits (int, optional): Length of keys. Defaults to 1024.
            format (Literal[, optional): Output format type. Defaults to 'PEM'.
            passphrase (str, optional): Passphrase for keys. Defaults to None.

        Returns:
            Chepy: The Chepy object.
        """
        key = RSA.generate(bits)
        pub = key.public_key().exportKey(passphrase=passphrase, format=format)
        priv = key.exportKey(passphrase=passphrase, format=format)
        self.state = {"public": pub, "private": priv}
        return self

    @ChepyDecorators.call_stack
    def generate_ecc_keypair(
        self,
        curve: Literal["p256", "p384", "p521"] = "p256",
        format: Literal["PEM", "DER"] = "PEM",
    ) -> PublickeyT:
        """Generate RSA key pair

        Args:
            curve (Literal[, optional): Curve for keys. Defaults to p256.
            format (Literal[, optional): Output format type. Defaults to 'PEM'.

        Returns:
            Chepy: The Chepy object.
        """
        key = ECC.generate(curve=curve)
        pub = key.public_key().export_key(format=format)
        priv = key.export_key(format=format)
        self.state = {"public": pub, "private": priv}
        return self



from typing import Any, Dict, List

class Encoding:
    py_encodings: Any = ...
    py_text_encodings: Any = ...
    asciichars: Any = ...
    brailles: Any = ...
    wingdings: dict = ...
    NATO_CONSTANTS_DICT: Dict[str, str] = ...
    LEETCODE: List[List[str]] = ...
    BASE91_ALPHABET: List[str] = ...

class EncryptionConsts:
    MORSE_CODE_DICT: Any = ...

class PcapUSB:
    qwerty_map: Any = ...
    qwerty_modifier: Any = ...
    dvorak: Any = ...
    dvorak_modifier: Any = ...



from ..core import ChepyCore
from typing import Any, TypeVar, Literal

CompressionT = TypeVar('CompressionT', bound='Compression')

class Compression(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def fix_zip_header(self: CompressionT) -> CompressionT: ...
    def zip_info(self: CompressionT) -> CompressionT: ...
    def zip_list_files(self: CompressionT) -> CompressionT: ...
    def unzip_one(self: CompressionT, file: str, password: str=...) -> CompressionT: ...
    def unzip_all(self: CompressionT, password: str=...) -> CompressionT: ...
    def create_zip_file(self: CompressionT, file_name: str) -> CompressionT: ...
    def zip_compress(self: CompressionT, file_name: str) -> CompressionT: ...
    def zip_compress_symlink(self: CompressionT, file_name: str, target_file: str) -> CompressionT: ...
    def tar_list_files(self: CompressionT, mode: Literal["gz", "bz2", "xz", ""]=...) -> CompressionT: ...
    def tar_extract_one(self: CompressionT, filename: str, mode: Literal["gz", "bz2", "xz", ""]=...) -> CompressionT: ...
    def tar_extract_all(self: CompressionT, mode: Literal["gz", "bz2", "xz", ""]=...) -> CompressionT: ...
    def tar_compress(self: CompressionT, filename: str, mode: Literal["gz", "bz2", "xz", ""]=...) -> CompressionT: ...
    def gzip_compress(self: CompressionT, file_name: str=...) -> CompressionT: ...
    def gzip_decompress(self: CompressionT) -> CompressionT: ...
    def bzip_compress(self: CompressionT) -> CompressionT: ...
    def bzip_decompress(self: CompressionT) -> CompressionT: ...
    def zlib_compress(self: CompressionT, level: int=...) -> CompressionT: ...
    def zlib_decompress(self: CompressionT) -> CompressionT: ...
    def lzma_compress(self: CompressionT) -> CompressionT: ...
    def lzma_decompress(self: CompressionT) -> CompressionT: ...
    def raw_inflate(self: CompressionT) -> CompressionT: ...
    def raw_deflate(self: CompressionT) -> CompressionT: ...
    def lz4_compress(self: CompressionT) -> CompressionT: ...
    def lz4_decompress(self: CompressionT) -> CompressionT: ...
    def lz77_compress(self: CompressionT, window_size: int = 13, lookahead_buffer_size: int = 6) -> CompressionT: ...
    def lz77_decompress(self: CompressionT, window_size: int = 13, lookahead_buffer_size: int = 6) -> CompressionT: ...



from prompt_toolkit.completion import Completer
from typing import Any

module: Any
options: Any
config: Any

class CliCompleter(Completer):
    def get_completions(self, document: Any, complete_event: Any) -> None: ...

def functions_cli(): ...
def get_doc(method: str) -> Any: ...
def cli_edit_state(fire: object, args: list) -> Any: ...
def cli_highlight(fire: object, highlight: str) -> Any: ...
def get_cli_options(): ...
def print_in_colors(out: Any) -> None: ...
def cli_state_type(fire: object) -> Any: ...
def cli_get_state(fire: object, index: int) -> Any: ...
def cli_show_length(fire: object) -> Any: ...
def cli_show_dict_keys(fire: object, pretty: bool=...) -> Any: ...
def cli_show_states(fire: object, pretty: bool=...) -> Any: ...
def cli_show_buffers(fire: object, pretty: bool=...) -> Any: ...
def cli_get_attr(fire: object, attr: str) -> Any: ...
def cli_pretty_print(fire: object) -> Any: ...
def cli_plugin_path(config: Any) -> None: ...
def cli_show_errors(errors: Any) -> None: ...
def cli_go_back() -> None: ...
def cli_delete_history() -> None: ...
def cli_exit(fire: object) -> Any: ...



from ..core import ChepyCore
from typing import Any, Literal, TypeVar, Union, List

yaml: Any
DataFormatT = TypeVar('DataFormatT', bound='DataFormat')

class DataFormat(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def eval_state(self: DataFormatT) -> DataFormatT: ...
    def list_to_bytes(self: DataFormatT, ascii: bool=False) -> DataFormatT: ...
    def list_to_str(self: DataFormatT, join_by: Union[str, bytes]=...) -> DataFormatT: ...
    def str_list_to_list(self: DataFormatT) -> DataFormatT: ...
    def join(self: DataFormatT, join_by: Union[str, bytes]=...) -> DataFormatT: ...
    def json_to_dict(self: DataFormatT) -> DataFormatT: ...
    def dict_to_json(self: DataFormatT) -> DataFormatT: ...
    def dict_get_items(self: DataFormatT, *keys: str) -> DataFormatT: ...
    def yaml_to_json(self: DataFormatT) -> DataFormatT: ...
    def json_to_yaml(self: DataFormatT) -> DataFormatT: ...
    def to_base58(self: DataFormatT) -> DataFormatT: ...
    def from_base58(self: DataFormatT) -> DataFormatT: ...
    def to_base85(self: DataFormatT) -> DataFormatT: ...
    def from_base85(self: DataFormatT) -> DataFormatT: ...
    def to_base16(self: DataFormatT) -> DataFormatT: ...
    def from_base16(self: DataFormatT) -> DataFormatT: ...
    def to_base32(self: DataFormatT) -> DataFormatT: ...
    def from_base32(self: DataFormatT, remove_whitespace: bool=True) -> DataFormatT: ...
    def to_int(self: DataFormatT, byteorder: Literal['big', 'little']='big', base: int=10) -> DataFormatT: ...
    def to_bytes(self: DataFormatT) -> DataFormatT: ...
    def from_bytes(self: DataFormatT) -> DataFormatT: ...
    def to_base64(self: DataFormatT, alphabet: Literal[str, 'standard', 'url_safe', 'filename_safe', 'itoa64', 'xml', 'z64', 'radix_64', 'xxencoding', 'rot13', 'unix_crypt']='standard') -> DataFormatT: ...
    def from_base64(self: DataFormatT, alphabet: Literal[str, 'standard', 'url_safe', 'filename_safe', 'itoa64', 'xml', 'z64', 'radix_64', 'xxencoding', 'rot13', 'unix_crypt']='standard', remove_non_alpha: bool=True) -> DataFormatT: ...
    def decode_bytes(self: DataFormatT, errors: Literal['ignore', 'backslashreplace', 'replace']=...) -> DataFormatT: ...
    def to_hex(self: DataFormatT, delimiter: str=..., join_by: str=...) -> DataFormatT: ...
    def from_hex(self: DataFormatT, delimiter: Union[str, None]=None, join_by: str='', replace: Union[bytes, None]=b'%|0x') -> DataFormatT: ...
    def hex_to_int(self: DataFormatT) -> DataFormatT: ...
    def hex_to_bytes(self: DataFormatT) -> DataFormatT: ...
    def hex_to_str(self: DataFormatT, ignore: bool=...) -> DataFormatT: ...
    def str_to_hex(self: DataFormatT, delimiter: Union[str, bytes]=b'') -> DataFormatT: ...
    def int_to_bytes(self: DataFormatT, length: Union[int, None], byteorder: Literal['big,little']='big') -> DataFormatT: ...
    def int_to_hex(self: DataFormatT, unsigned_int: bool=True, bit_size: int=64) -> DataFormatT: ...
    def int_to_str(self: DataFormatT) -> DataFormatT: ...
    def binary_to_hex(self: DataFormatT) -> DataFormatT: ...
    def normalize_hex(self: DataFormatT, is_bytearray: Any=...) -> DataFormatT: ...
    def str_from_hexdump(self: DataFormatT) -> DataFormatT: ...
    def to_hexdump(self: DataFormatT) -> DataFormatT: ...
    def from_hexdump(self: DataFormatT) -> DataFormatT: ...
    def from_url_encoding(self: DataFormatT) -> DataFormatT: ...
    def to_url_encoding(self: DataFormatT, safe: str=..., all_chars: bool=...) -> DataFormatT: ...
    def bytearray_to_str(self: DataFormatT, encoding: str=..., errors: str=...) -> DataFormatT: ...
    def to_list(self: DataFormatT) -> DataFormatT: ...
    def str_to_dict(self: DataFormatT) -> DataFormatT: ...
    def to_charcode(self: DataFormatT, join_by: str=..., base: int=...) -> DataFormatT: ...
    def from_charcode(self: DataFormatT, delimiter: Union[str, None]=None, join_by: str=' ', base: int=10) -> DataFormatT: ...
    def to_decimal(self: DataFormatT) -> DataFormatT: ...
    def from_decimal(self: DataFormatT, delimiter: Union[str, None]=None, join_by: str=...) -> DataFormatT: ...
    def to_binary(self: DataFormatT, join_by: Union[bytes, str]=..., byte_length: int=...) -> DataFormatT: ...
    def from_binary(self: DataFormatT, delimiter: Union[str, None]=None, byte_length: int=...) -> DataFormatT: ...
    def to_octal(self: DataFormatT, join_by: str=...) -> DataFormatT: ...
    def from_octal(self: DataFormatT, delimiter: str=..., join_by: str=...) -> DataFormatT: ...
    def to_html_entity(self: DataFormatT, format:Literal['named', 'numeric', 'hex']='named', all_chars:bool=False) -> DataFormatT: ...
    def from_html_entity(self: DataFormatT) -> DataFormatT: ...
    def to_punycode(self: DataFormatT) -> DataFormatT: ...
    def from_punycode(self: DataFormatT) -> DataFormatT: ...
    def encode_bruteforce(self: DataFormatT) -> DataFormatT: ...
    def decode_bruteforce(self: DataFormatT) -> DataFormatT: ...
    def to_braille(self: DataFormatT) -> DataFormatT: ...
    def from_braille(self: DataFormatT) -> DataFormatT: ...
    def trim(self: DataFormatT) -> DataFormatT: ...
    def to_nato(self: DataFormatT, join_by:str=...) -> DataFormatT: ...
    def from_nato(self: DataFormatT, delimiter: Union[str, None]=None, join_by: str=...) -> DataFormatT: ...
    def swap_values(self: DataFormatT, indices1: Union[str, List[int]], indices2: Union[str, List[int]]) -> DataFormatT: ...
    def swap_strings(self: DataFormatT, by:int) -> DataFormatT: ...
    def to_string(self: DataFormatT) -> DataFormatT: ...
    def stringify(self: DataFormatT, compact:bool=...) -> DataFormatT: ...
    def select(self: DataFormatT, start: Union[int, str, bytes], end: Union[None, int]=None) -> DataFormatT: ...
    def length(self: DataFormatT) -> DataFormatT: ...
    def to_leetcode(self: DataFormatT, replace_space: str=...) -> DataFormatT: ...
    def substitute(self: DataFormatT, x: str=..., y: str=...) -> DataFormatT: ...
    def remove(self: DataFormatT, pattern: Union[str, bytes] = ...): ...
    def remove_nonprintable(self: DataFormatT, replace_with: Union[str, bytes] = ...): ...
    def to_base92(self: DataFormatT) -> DataFormatT: ...
    def from_base92(self: DataFormatT) -> DataFormatT: ...
    def to_base45(self: DataFormatT) -> DataFormatT: ...
    def from_base45(self: DataFormatT) -> DataFormatT: ...
    def to_base91(self: DataFormatT) -> DataFormatT: ...
    def from_base91(self: DataFormatT) -> DataFormatT: ...
    def swap_endianness(self: DataFormatT, word_length: int=4, pad_incomplete: bool=True) -> DataFormatT: ...
    def bruteforce_from_base_xx(self: DataFormatT) -> DataFormatT: ...
    def long_to_bytes(self: DataFormatT) -> DataFormatT: ...
    def bytes_to_long(self: DataFormatT) -> DataFormatT: ...
    def concat(self: DataFormatT, data: Union[str, bytes]) -> DataFormatT: ...
    def to_wingdings(self: DataFormatT) -> DataFormatT: ...
    def from_wingdings(self: DataFormatT) -> DataFormatT: ...
    def to_twin_hex(self: DataFormatT) -> DataFormatT: ...
    def from_twin_hex(self: DataFormatT) -> DataFormatT: ...
    def to_base36(self: DataFormatT, join_by: Union[str, bytes]=...) -> DataFormatT: ...
    def from_base36(self: DataFormatT, delimiter: Union[str, bytes]=..., join_by: Union[str, bytes]=...) -> DataFormatT: ...
    def to_pickle(self: DataFormatT) -> DataFormatT: ...
    def from_pickle(self: DataFormatT, trust: bool=...) -> DataFormatT: ...
    def to_bacon(self: DataFormatT, A: Literal['A', '0']=..., B: Literal['B', '1']=..., complete: bool=..., join_by: bytes=..., invert: bool=...) -> DataFormatT: ...
    def from_bacon(self: DataFormatT, A: Literal['A', '0']=..., B: Literal['B', '1']=..., complete: bool=..., split_by: bytes=..., invert: bool=...) -> DataFormatT: ...
    def to_upside_down(self: DataFormatT, reverse: bool=False) -> DataFormatT: ...
    def from_upside_down(self: DataFormatT, reverse: bool=False) -> DataFormatT: ...
    def to_messagepack(self: DataFormatT) -> DataFormatT: ...
    def from_messagepack(self: DataFormatT) -> DataFormatT: ...
    def unicode_escape(self: DataFormatT, padding: int = 0, uppercase_hex:bool=False) -> DataFormatT: ...
    def to_base(self: DataFormatT, radix: int=16) -> DataFormatT: ...
    def from_base(self: DataFormatT, radix: int=16) -> DataFormatT: ...
    def rotate_right(self: DataFormatT, radix: int=1, carry: bool=False) -> DataFormatT: ...
    def rotate_left(self: DataFormatT, radix: int=1, carry: bool=False) -> DataFormatT: ...
    def to_base62(self: DataFormatT, alphabet: str="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz") -> DataFormatT: ...
    def from_base62(self: DataFormatT, alphabet: str="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz") -> DataFormatT: ...
    def to_base65536(self: DataFormatT) -> DataFormatT: ...
    def from_base65536(self: DataFormatT) -> DataFormatT: ...
    def cut(self: DataFormatT, start: int, end: int) -> DataFormatT: ...
    def flatten(self: DataFormatT) -> DataFormatT: ...
    def to_utf21(self: DataFormatT) -> DataFormatT: ...
    def from_utf21(self: DataFormatT) -> DataFormatT: ...
    def to_uuencode(self: DataFormatT, header: str='-') -> DataFormatT: ...
    def from_uuencode(self: DataFormatT, header: str='-') -> DataFormatT: ...
    def from_quoted_printable(self: DataFormatT) -> DataFormatT: ...
    def to_quoted_printable(self: DataFormatT) -> DataFormatT: ...
    def from_rison(self: DataFormatT) -> DataFormatT: ...
    def to_rison(self: DataFormatT) -> DataFormatT: ...
    def increment_bytes(self: DataFormatT, n: int) -> DataFormatT: ...
    def decrement_bytes(self: DataFormatT, n: int) -> DataFormatT: ...
    def parse_csv(self: DataFormatT) -> DataFormatT: ...
    def parse_sqlite(self: DataFormatT, query: str) -> DataFormatT: ...
    def to_italics(self: DataFormatT) -> DataFormatT: ...
    def from_italics(self: DataFormatT) -> DataFormatT: ...
    def to_fullwidth(self: DataFormatT) -> DataFormatT: ...
    def from_fullwidth(self: DataFormatT) -> DataFormatT: ...



from ..core import ChepyCore
from typing import Any, Literal, TypeVar, Union

RSA: Any
OpenSSL: Any
PublickeyT = TypeVar('PublickeyT', bound='Publickey')

class Publickey(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def parse_x509_pem(self: PublickeyT) -> PublickeyT: ...
    def parse_x509_der_hex(self: PublickeyT) -> PublickeyT: ...
    def public_from_x509(self: PublickeyT) -> PublickeyT: ...
    def pem_to_der_hex(self: PublickeyT) -> PublickeyT: ...
    def der_hex_to_pem(self: PublickeyT) -> PublickeyT: ...
    def parse_public_pem(self: PublickeyT) -> PublickeyT: ...
    def parse_private_pem(self: PublickeyT) -> PublickeyT: ...
    def dump_pkcs12_cert(self: PublickeyT, password: Union[str, bytes]) -> PublickeyT: ...
    def generate_rsa_keypair(self: PublickeyT, bits:int=..., format:str=..., passphrase:str=...) -> PublickeyT: ...
    def generate_ecc_keypair(self: PublickeyT, curve:Literal['p256', 'p384', 'p521']='p256', format:Literal['PEM', 'DER']='PEM') -> PublickeyT: ...
